"""Define a cell object with input are and rich outputs, and related objects."""

from __future__ import annotations

import asyncio
import logging
import os
import weakref
from functools import lru_cache, partial
from pathlib import Path
from typing import TYPE_CHECKING, cast
from weakref import WeakKeyDictionary

import nbformat
from prompt_toolkit.completion.base import (
    DynamicCompleter,
    _MergedCompleter,
)
from prompt_toolkit.document import Document
from prompt_toolkit.filters.app import is_searching
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.utils import Event

from euporie.core.app.current import get_app
from euporie.core.border import NoLine, ThickLine, ThinLine
from euporie.core.completion import DeduplicateCompleter, LspCompleter
from euporie.core.diagnostics import Report
from euporie.core.filters import multiple_cells_selected
from euporie.core.format import LspFormatter
from euporie.core.inspection import (
    FirstInspector,
    LspInspector,
)
from euporie.core.layout.containers import HSplit, VSplit, Window
from euporie.core.lsp import LspCell
from euporie.core.utils import on_click
from euporie.core.widgets.cell_outputs import CellOutputArea
from euporie.core.widgets.inputs import KernelInput, StdInput

if TYPE_CHECKING:
    from typing import Any, Callable, Literal

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.completion.base import Completer
    from prompt_toolkit.formatted_text.base import StyleAndTextTuples

    from euporie.core.border import GridStyle
    from euporie.core.format import Formatter
    from euporie.core.inspection import Inspector
    from euporie.core.lsp import LspClient
    from euporie.core.tabs.notebook import BaseNotebook


log = logging.getLogger(__name__)


def get_cell_id(cell_json: dict) -> str:
    """Return the cell ID field defined in a cell JSON object.

    If no cell ID is defined (as per ```:mod:`nbformat`<4.5``), then one is generated
    and added to the cell.

    Args:
        cell_json: The cell's JSON object as a python dictionary

    Returns:
        The ID string

    """
    cell_id = cell_json.get("id", "")
    # Assign a cell id if missing
    if not cell_id:
        cell_json["id"] = cell_id = nbformat.v4.new_code_cell().get("id")
    return cell_id


@lru_cache(maxsize=32)
def _get_border_style(
    selected: bool, focused: bool, show_borders: bool, multi_selected: bool
) -> GridStyle:
    """Get the border style grid based on cell state."""
    if not (show_borders or selected):
        return NoLine.grid
    if focused and multi_selected:
        return ThickLine.outer
    return ThinLine.outer


class Cell:
    """A kernel_tab cell element.

    Contains a transparent clickable overlay, which is not displayed when the cell is
    focused.
    """

    def __init__(
        self, index: int, json: dict, kernel_tab: BaseNotebook, is_new: bool = False
    ) -> None:
        """Initiate the cell element.

        Args:
            index: The position of this cell in the kernel_tab
            json: A reference to the cell's json object
            kernel_tab: The kernel_tab instance this cell belongs to
            is_new: Flag a newly inserted cell

        """
        weak_self = weakref.proxy(self)
        self.index = index
        self.json = json
        self.kernel_tab: BaseNotebook = kernel_tab
        self.rendered = True
        self.clear_outputs_on_output = False
        self.state = "idle"
        self.is_new = is_new

        self.on_open = Event(self)
        self.on_change = Event(self)
        self.on_close = Event(self)

        self.inspectors: list[Inspector] = []
        self.inspector = FirstInspector(
            lambda: [*self.kernel_tab.inspectors, *self.inspectors]
        )
        self.completers: list[Completer] = []
        self.completer = DeduplicateCompleter(
            DynamicCompleter(
                lambda: _MergedCompleter(
                    [*self.kernel_tab.completers, *self.completers]
                )
            )
        )
        self.formatters: list[Formatter] = [*self.kernel_tab.formatters]
        self.reports: WeakKeyDictionary[LspClient, Report] = WeakKeyDictionary()

        show_input = Condition(
            lambda: bool(
                (weak_self.json.get("cell_type") != "markdown")
                | (
                    (weak_self.json.get("cell_type") == "markdown")
                    & ~weak_self.rendered
                )
            )
        )
        show_output = Condition(
            lambda: (
                (weak_self.json.get("cell_type") != "markdown")
                & bool(weak_self.output_json)
                | ((weak_self.json.get("cell_type") == "markdown") & weak_self.rendered)
            )
        )
        # scroll_input = Condition(
        # lambda: bool(
        # (weak_self.json.get("cell_type") == "markdown") & ~weak_self.rendered
        # )
        # )
        show_prompt = Condition(lambda: weak_self.cell_type == "code")
        self.is_code = Condition(lambda: weak_self.json.get("cell_type") == "code")

        self.output_area = CellOutputArea(json=self.output_json, parent=weak_self)

        def on_text_changed(buf: Buffer) -> None:
            """Update cell json when the input buffer has been edited."""
            weak_self._set_input(buf.text)
            weak_self.kernel_tab.dirty = True
            weak_self.on_change()
            # Re-render markdown cells when edited outside of edit mode
            if (
                weak_self.cell_type == "markdown"
                and not weak_self.kernel_tab.in_edit_mode()
            ):
                weak_self.output_area.json = weak_self.output_json
                weak_self.refresh()

        def on_cursor_position_changed(buf: Buffer) -> None:
            """Respond to cursor movements."""
            # Tell the scrolling container to scroll the cursor into view on the next render
            weak_self.kernel_tab.page.scroll_to_cursor = True
            if not is_searching():
                if not self.selected:
                    weak_self.kernel_tab.select(self.index, scroll=True)
                if not weak_self.kernel_tab.in_edit_mode():
                    weak_self.kernel_tab.enter_edit_mode()

        # Now we generate the main container used to represent a kernel_tab cell

        source_hidden = Condition(
            lambda: weak_self.json.get("metadata", {})
            .get("jupyter", {})
            .get("source_hidden", False)
        )

        self.input_box = KernelInput(
            kernel_tab=self.kernel_tab,
            text=self.input,
            completer=self.completer,
            complete_while_typing=self.is_code,
            autosuggest_while_typing=self.is_code,
            wrap_lines=Condition(lambda: weak_self.json.get("cell_type") == "markdown"),
            on_text_changed=on_text_changed,
            on_cursor_position_changed=on_cursor_position_changed,
            language=partial(lambda cell: cell.language, weakref.proxy(self)),
            accept_handler=lambda buffer: weak_self.run_or_render() or True,
            focusable=show_input & ~source_hidden,
            tempfile_suffix=self.suffix,
            inspector=self.inspector,
            diagnostics=self.report,
            formatters=self.formatters,
            # show_diagnostics=Condition(
            #     lambda: kernel_tab.app.layout.has_focus(self.input_box.buffer)
            # ),
        )
        self.input_box.buffer.name = self.cell_type

        def border_char(name: str) -> Callable[..., str]:
            """Return a function which returns the cell border character to display."""

            def _inner() -> str:
                if not weak_self:
                    return " "
                grid = _get_border_style(
                    weak_self.selected,
                    weak_self.focused,
                    weak_self.kernel_tab.app.config.show_cell_borders,
                    multiple_cells_selected(),
                )
                return getattr(grid, name.upper())

            return _inner

        self.control = Window(
            FormattedTextControl(
                border_char("TOP_LEFT"),
                focusable=True,
                show_cursor=False,
            ),
            width=1,
            height=0,
            style="class:border",
            always_hide_cursor=True,
        )

        fill = partial(Window, style="class:border")

        # Create textbox for standard input
        self.stdin_box = StdInput(weak_self.kernel_tab)

        input_row = ConditionalContainer(
            VSplit(
                [
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(
                                lambda: [
                                    (
                                        "",
                                        weak_self.prompt,
                                        on_click(self.toggle_input),
                                    ),
                                    ("", "\n ", lambda e: NotImplemented),
                                ]
                            ),
                            width=lambda: len(weak_self.prompt) + 1,
                            height=Dimension(preferred=1),
                            style="class:input,prompt",
                        ),
                        filter=show_prompt,
                    ),
                    ConditionalContainer(self.input_box, filter=~source_hidden),
                    ConditionalContainer(
                        Window(
                            FormattedTextControl(
                                cast(
                                    "StyleAndTextTuples",
                                    [
                                        (
                                            "class:cell,show,inputs,border",
                                            "▏",
                                            on_click(self.show_input),
                                        ),
                                        (
                                            "class:cell,show,inputs",
                                            "…",
                                            on_click(self.show_input),
                                        ),
                                        (
                                            "class:cell,show,inputs,border",
                                            "▕",
                                            on_click(self.show_input),
                                        ),
                                    ],
                                )
                            )
                        ),
                        filter=source_hidden,
                    ),
                ],
            ),
            filter=show_input,
        )

        outputs_hidden = Condition(
            lambda: weak_self.json["metadata"]
            .get("jupyter", {})
            .get("outputs_hidden", False)
            or weak_self.json["metadata"].get("collapsed", False)
        )

        output_row = ConditionalContainer(
            VSplit(
                [
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(
                                lambda: [
                                    (
                                        "",
                                        weak_self.prompt,
                                        on_click(self.toggle_output),
                                    ),
                                    ("", "\n ", lambda e: NotImplemented),
                                ],
                            ),
                            width=lambda: len(weak_self.prompt) + 1,
                            height=Dimension(preferred=1),
                            style="class:output,prompt",
                        ),
                        filter=show_prompt,
                    ),
                    HSplit(
                        [
                            ConditionalContainer(
                                self.output_area,
                                filter=~outputs_hidden,
                            ),
                            ConditionalContainer(
                                Window(
                                    FormattedTextControl(
                                        cast(
                                            "StyleAndTextTuples",
                                            [
                                                (
                                                    "class:cell,show,outputs,border",
                                                    "▏",
                                                    on_click(self.show_output),
                                                ),
                                                (
                                                    "class:cell,show,outputs",
                                                    "…",
                                                    on_click(self.show_output),
                                                ),
                                                (
                                                    "class:cell,show,outputs,border",
                                                    "▕",
                                                    on_click(self.show_output),
                                                ),
                                            ],
                                        )
                                    )
                                ),
                                filter=outputs_hidden,
                            ),
                            self.stdin_box,
                        ],
                    ),
                ],
                width=Dimension(min=1),
            ),
            filter=show_output | self.stdin_box.visible,
        )

        def _style() -> str:
            """Calculate the cell's style given its state."""
            style = "class:cell"
            if weak_self.selected:
                if weak_self.kernel_tab.in_edit_mode():
                    style += ",edit"
                else:
                    style += ",cell.selection"
            return style

        self.container = HSplit(
            [
                VSplit(
                    [
                        self.control,
                        fill(char=border_char("TOP_MID"), height=1),
                        fill(width=1, height=1, char=border_char("TOP_RIGHT")),
                    ],
                    height=1,
                ),
                VSplit(
                    [
                        fill(width=1, char=border_char("MID_LEFT")),
                        HSplit(
                            [input_row, output_row],
                            padding=lambda: 1 if show_input() and show_output() else 0,
                        ),
                        fill(width=1, char=border_char("MID_RIGHT")),
                    ]
                ),
                VSplit(
                    [
                        fill(width=1, height=1, char=border_char("BOTTOM_LEFT")),
                        fill(char=border_char("BOTTOM_MID")),
                        fill(width=1, height=1, char=border_char("BOTTOM_RIGHT")),
                    ],
                    height=1,
                ),
            ],
            style=_style,
        )

        self.kernel_tab.app.create_background_task(self.load_lsps())

    async def load_lsps(self) -> None:
        """Add hooks to the notebook LSP client."""
        path = self.path
        weak_self = weakref.proxy(self)

        # Wait for all lsps to be initialized, and setup hooks as they become ready
        async def _await_load(lsp: LspClient) -> LspClient:
            await lsp.initialized.wait()
            return lsp

        for ready in asyncio.as_completed(
            [_await_load(lsp) for lsp in self.kernel_tab.lsps]
        ):
            lsp = await ready

            # Listen for LSP diagnostics
            lsp.on_diagnostics += weak_self.lsp_update_diagnostics

            open_handler = partial(lambda lsp, tab: self.lsp_open_handler(lsp), lsp)
            change_handler = partial(lambda lsp, tab: self.lsp_change_handler(lsp), lsp)
            close_handler = partial(lambda lsp, tab: self.lsp_close_handler(lsp), lsp)

            self.on_open += open_handler
            self.on_change += change_handler
            self.on_close += close_handler

            # Add completer
            completer = LspCompleter(lsp, path)
            self.completers.append(completer)

            # Add inspector
            inspector = LspInspector(lsp, path)
            self.inspectors.append(inspector)

            # Add formatter
            formatter = LspFormatter(lsp, path)
            self.formatters.append(formatter)

            def lsp_unload(lsp: LspClient) -> None:
                self.on_open -= open_handler  # noqa: B023
                self.on_change -= change_handler  # noqa: B023
                self.on_close -= close_handler  # noqa: B023
                if completer in weak_self.completers:  # noqa: B023
                    self.completers.remove(completer)  # noqa: B023
                if inspector in weak_self.inspectors:  # noqa: B023
                    self.inspectors.remove(inspector)  # noqa: B023
                if formatter in weak_self.formatters:  # noqa: B023
                    self.formatters.remove(formatter)  # noqa: B023

            lsp.on_exit += lsp_unload

            if self.is_new:
                self.on_open()

    @property
    def lsp_cell(self) -> LspCell:
        """Capture the cell's attributes for an LSP server."""
        return LspCell(
            id=self.id,
            idx=self.index,
            path=self.path,
            kind=self.cell_type,
            language=self.language,
            execution_count=self.execution_count,
            metadata=self.json["metadata"],
            text=self.input,
        )

    def lsp_open_handler(self, lsp: LspClient) -> None:
        """If the LSP does not support notebooks, open this cell as a text document."""
        if lsp.can_open_nb:
            lsp.change_nb_add(path=self.kernel_tab.path, cells=[self.lsp_cell])
        else:
            lsp.open_doc(
                path=self.path,
                language=self.language,
                text=self.input_box.buffer.text,
            )

    def lsp_change_handler(self, lsp: LspClient) -> None:
        """Tell the LSP server a cell has changed."""
        if lsp.can_change_nb:
            lsp.change_nb_edit(path=self.kernel_tab.path, cells=[self.lsp_cell])
        else:
            lsp.change_doc(
                path=self.path,
                language=self.language,
                text=self.input_box.buffer.text,
            )

    def lsp_close_handler(self, lsp: LspClient) -> None:
        """Notify the LSP of a deleted cell."""
        if lsp.can_change_nb:
            lsp.change_nb_delete(path=self.kernel_tab.path, cells=[self.lsp_cell])
        else:
            lsp.close_doc(self.path)

    def lsp_update_diagnostics(self, lsp: LspClient) -> None:
        """Process a new diagnostic report from the LSP."""
        if (diagnostics := lsp.reports.pop(self.path.as_uri(), None)) is not None:
            self.reports[lsp] = Report.from_lsp(self.input_box.text, diagnostics)
            self.refresh()

    def report(self) -> Report:
        """Return the current diagnostic reports."""
        return Report.from_reports(*self.reports.values())

    def focus(self, position: int | None = None, scroll: bool = False) -> None:
        """Focus the relevant control in this cell.

        Args:
            position: An optional cursor position index to apply to the input box
            scroll: Whether to scroll the page to make the selection visible

        """
        to_focus = None
        if self.kernel_tab.edit_mode:
            # Select just this cell when editing
            # self.kernel_tab.select(self.index)
            if self.stdin_box.visible():
                to_focus = self.stdin_box.window
            else:
                self.show_input()
                to_focus = self.input_box.window
                self.rendered = False
            if position is not None:
                self.input_box.buffer.cursor_position = position % (
                    len(self.input_box.buffer.text) + 1
                )
        else:
            to_focus = self.kernel_tab.cell.control

        # We force focus here, bypassing the layout's checks, as the control we want to
        # focus might be not be in the current layout yet.
        get_app().layout._stack.append(to_focus)

        # Scroll the currently selected slice into view
        if scroll:
            self.kernel_tab.scroll_to(self.index)

    @property
    def focused(self) -> bool:
        """Determine if the cell currently has focus."""
        if self.container is not None:
            return get_app().layout.has_focus(self)
        else:
            return False

    @property
    def selected(self) -> bool:
        """Determine if the cell currently is selected."""
        if self.container is not None:
            return self.index in self.kernel_tab.selected_indices
        else:
            return False

    @property
    def cell_type(self) -> str:
        """Determine the current cell type."""
        return self.json.get("cell_type", "code")

    def suffix(self) -> str:
        """Return the file suffix matching the current cell type."""
        cell_type = self.cell_type
        if cell_type == "markdown":
            return ".md"
        elif cell_type == "raw":
            return ""
        else:
            return self.kernel_tab.kernel_lang_file_ext

    @property
    def execution_count(self) -> int:
        """Retrieve the execution count from the cell's JSON."""
        return self.json.get("execution_count", 0)

    @execution_count.setter
    def execution_count(self, count: int) -> None:
        """Set the execution count in the cell's JSON.

        Args:
            count: The new execution count number.

        """
        self.json["execution_count"] = count

    @property
    def prompt(self) -> str:
        """Determine what should be displayed in the prompt of the cell."""
        if self.state in ("busy", "queued"):
            prompt = "*"
        else:
            prompt = str(self.execution_count or " ")
        if prompt:
            prompt = f"[{prompt}]"
        return prompt

    def _set_input(self, value: str) -> None:
        self.json["source"] = value

    @property
    def input(self) -> str:
        """Fetch the cell's contents from the cell's JSON."""
        return self.json.get("source", "")

    @input.setter
    def input(self, value: str) -> None:
        """Set the cell's contents in the cell's JSON.

        Args:
            value: The new cell contents text.

        """
        cp = self.input_box.buffer.cursor_position
        self._set_input(value)
        self.input_box.text = self.json["source"]
        cp = max(0, min(cp, len(value)))
        self.input_box.buffer.cursor_position = cp

    @property
    def output_json(self) -> list[dict[str, Any]]:
        """Retrieve a list of cell outputs from the cell's JSON."""
        if self.cell_type == "markdown":
            return [
                {"data": {"text/x-markdown": self.input}, "output_type": "markdown"}
            ]
        else:
            return self.json.setdefault("outputs", [])

    def refresh(self, now: bool = True) -> None:
        """Request that the cell to be re-rendered next time it is drawn."""
        self.kernel_tab.refresh_cell(self)
        if now:
            self.kernel_tab.app.invalidate()

    def ran(self, content: dict | None = None) -> None:
        """Update the cell status and update display when the cell has finished."""
        self.state = "idle"
        self.refresh()

    def remove_outputs(self) -> None:
        """Remove all outputs from the cell."""
        self.clear_outputs_on_output = False
        if "outputs" in self.json:
            self.json["outputs"].clear()
        if self.cell_type != "markdown":
            self.output_area.reset()
        # Ensure the cell output area is visible
        self.show_output()
        self.refresh()

    def set_cell_type(
        self, cell_type: Literal["markdown", "code", "raw"], clear: bool = False
    ) -> None:
        """Convert the cell to a different cell type.

        Args:
            cell_type: The desired cell type.
            clear: If True, cell outputs will be cleared

        """
        if clear and cell_type != "markdown":
            self.remove_outputs()
        if cell_type == "code":
            self.json.setdefault("execution_count", None)
        if cell_type == "markdown" and "execution_count" in self.json:
            del self.json["execution_count"]
        # Record the new cell type
        self.json["cell_type"] = cell_type
        self.input_box.buffer.name = cell_type
        # Update the output-area
        # self.output_area.json = self.output_json
        # Force the input box lexer to re-run
        self.input_box.control._fragment_cache.clear()
        # Trigger callbacks
        self.on_change()
        # Flag notebook as modified
        self.kernel_tab.dirty = True

    @property
    def path(self) -> Path:
        """Return a virtual path for this cell (used by LSP clients).

        Pylance appears to use URIs like:
        ``../folder.notebook.ipynb:pylance-notebook-cell:W0sZmlsZQ==.py``, which I've
        emulated here.
        """
        nb_path = self.kernel_tab.path
        return nb_path.parent / f"{nb_path.name}:cell:{self.id}{self.suffix}"

    @property
    def id(self) -> str:
        """Return the cell's ID as per the cell JSON."""
        return get_cell_id(self.json)

    @property
    def language(self) -> str:
        """Return the cell's code language."""
        if self.cell_type == "markdown":
            return "markdown"
        elif self.cell_type == "code":
            lang_info = self.kernel_tab.metadata.get("language_info", {})
            return lang_info.get("name", lang_info.get("pygments_lexer", "python"))
        else:
            return "raw"

    def run_or_render(
        self,
        buffer: Buffer | None = None,
        wait: bool = False,
        callback: Callable[..., None] | None = None,
    ) -> bool:
        """Send the cell's source code the the kernel to run.

        Args:
            buffer: Unused parameter, required when accepting the contents of a cell's
                input buffer
            wait: Has no effect
            callback: Callable to run when the kernel has finished running the cell

        Returns:
            Always returns True

        """
        if self.cell_type == "markdown":
            self.output_area.json = self.output_json
            self.rendered = True

        elif self.cell_type == "code":
            if get_app().config.autoformat:
                self.input_box.reformat()
            self.state = "queued"
            self.refresh()
            self.kernel_tab.run_cell(self, wait=wait, callback=callback)

        return True

    def __pt_container__(self) -> Container:
        """Return the container which represents this cell."""
        return self.container

    def set_execution_count(self, n: int) -> None:
        """Set the execution count of the cell."""
        self.json["execution_count"] = n
        self.refresh()

    def add_output(self, output_json: dict[str, Any], own: bool) -> None:
        """Add a new output to the cell."""
        # Clear the output if we were previously asked to
        if self.clear_outputs_on_output:
            self.remove_outputs()
        self.json.setdefault("outputs", []).append(output_json)
        # Add the new output to the output area
        self.output_area.add_output(output_json)
        # Tell the page this cell has been updated
        self.refresh()

    def clear_output(self, wait: bool = False) -> None:
        """Remove the cells output, optionally when new output is generated."""
        if wait:
            self.clear_outputs_on_output = True
        else:
            self.remove_outputs()

    def show_input(self) -> None:
        """Set the cell inputs to visible."""
        self.set_metadata(("jupyter", "source_hidden"), False)

    def hide_input(self) -> None:
        """Set the cell inputs to visible."""
        # Exit edit mode
        self.kernel_tab.edit_mode = False
        # Un-focus the cell input
        self.focus()
        # Set the input to hidden
        self.set_metadata(("jupyter", "source_hidden"), True)

    def toggle_input(self) -> None:
        """Toggle the visibility of the cell input."""
        if self.json["metadata"].get("jupyter", {}).get("source_hidden", False):
            self.show_input()
        else:
            self.hide_input()

    def show_output(self) -> None:
        """Set the cell outputs to visible."""
        self.set_metadata(("jupyter", "outputs_hidden"), False)
        self.set_metadata(("collapsed",), False)

    def hide_output(self) -> None:
        """Set the cell outputs to visible."""
        self.set_metadata(("jupyter", "outputs_hidden"), True)
        self.set_metadata(("collapsed",), True)

    def toggle_output(self) -> None:
        """Toggle the visibility of the cell outputs."""
        if self.json["metadata"].get("jupyter", {}).get("outputs_hidden", False):
            self.show_output()
        else:
            self.hide_output()

    def set_metadata(self, path: tuple[str, ...], data: Any) -> None:
        """Set a value in the metadata at an arbitrary path.

        Args:
            path: A tuple of path level names to create
            data: The value to add

        """
        level = self.json["metadata"]
        for i, key in enumerate(path):
            if i == len(path) - 1:
                level[key] = data
            else:
                level = level.setdefault(key, {})

    def set_status(self, status: str) -> None:
        """Set the execution status of the cell."""

    def get_input(
        self,
        prompt: str = "Please enter a value:",
        password: bool = False,
    ) -> None:
        """Scroll the cell requesting input into view and render it before asking for input."""
        self.kernel_tab.select(self.index)
        self.stdin_box.get_input(prompt, password)

    async def edit_in_editor(self) -> None:
        """Edit the cell in $EDITOR."""
        buffer = self.input_box.buffer
        app = get_app()
        edit_in_fg = False

        # Save VISUAL environment variable
        visual = os.environ.get("VISUAL")

        if editor := app.config.external_editor:
            if "{left}" in editor:
                win = self.input_box.window

                if (info := win.render_info) is not None:
                    edit_in_fg = True

                    margin_left = sum(
                        [win._get_margin_width(m) for m in win.left_margins]
                    )
                    margin_right = sum(
                        [win._get_margin_width(m) for m in win.right_margins]
                    )
                    top = info._y_offset
                    left = info._x_offset - margin_left
                    width = info.window_width + margin_left + margin_right
                    height = min(app.output.get_size().rows, info.window_height)

                else:
                    left = top = 0
                    height, width = app.output.get_size()

                editor = editor.format(
                    top=top,
                    left=left,
                    width=width,
                    height=height,
                    bottom=top + height,
                    right=left + width,
                )

            # Override VISUAL environment variable
            os.environ["VISUAL"] = editor

        if edit_in_fg:
            # Create a tempfile
            if buffer.tempfile:
                filename, cleanup_func = buffer._editor_complex_tempfile()
            else:
                filename, cleanup_func = buffer._editor_simple_tempfile()
            try:
                # Edit the temp file
                success = buffer._open_file_in_editor(filename)
                # Read content again.
                if success:
                    text = Path(filename).read_text()
                    # Drop trailing newline
                    if text.endswith("\n"):
                        text = text[:-1]
                    buffer.document = Document(text=text, cursor_position=len(text))
                    # Run the cell if configured
                    if app.config.run_after_external_edit:
                        buffer.validate_and_handle()
            finally:
                # Clean up temp dir/file.
                cleanup_func()

        else:
            await buffer.open_in_editor(
                validate_and_handle=app.config.run_after_external_edit
            )

        # Restore VISUAL environment variable
        if visual is not None:
            os.environ["VISUAL"] = visual

    def close(self) -> None:
        """Signal that the cell is no longer present in the notebook."""
        self.on_close()



"""Contain functions to automatically format code cell input."""

from __future__ import annotations

import logging
import subprocess
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from euporie.core.filters import command_exists
from euporie.core.lsp import range_to_slice

if TYPE_CHECKING:
    from pathlib import Path

    from euporie.core.lsp import LspClient


log = logging.getLogger(__name__)


class Formatter(metaclass=ABCMeta):
    """Text formatter class which reformats text."""

    def __init__(self, languages: set[str] | None = None) -> None:
        """Initialize a new formatter."""
        self.languages = languages or set()

    def _format(self, text: str, language: str) -> str:
        """Check the language matches, then format the input."""
        log.debug("Formatting using %s", self)
        if not self.languages or language in self.languages:
            return self.format(text)
        return text

    @abstractmethod
    def format(self, text: str) -> str:
        """Format the string."""
        return text


class CliFormatter(Formatter):
    """Format using an external command."""

    def __init__(self, command: list[str], languages: set[str] | None = None) -> None:
        """Create a new command formatter."""
        self.command = command
        self.filter = command_exists(command[0])
        super().__init__(languages=languages)

    def _format(self, text: str, language: str) -> str:
        """Check the filter value early."""
        if self.filter():
            return super()._format(text, language)
        return text

    def format(self, text: str) -> str:
        """Pass the text to the command over stdin and return the output."""
        try:
            proc = subprocess.Popen(
                self.command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                text=True,
            )
        except Exception:
            return text
        try:
            output, error = proc.communicate(text)
        except Exception:
            return text
        else:
            if output and not error:
                return output.rstrip("\r\n")
            else:
                return text

    def __repr__(self) -> str:
        """Return representation of the formatter as a string."""
        return f"{self.command[0].title()}Formatter()"


class LspFormatter(Formatter):
    """Format a document using a LSP server."""

    def __init__(
        self, lsp: LspClient, path: Path, languages: set[str] | None = None
    ) -> None:
        """Record the LSP client and file path for formatting."""
        self.lsp = lsp
        self.path = path
        super().__init__(languages=languages)

    def format(self, text: str) -> str:
        """Format the url with the LSP.

        The LSP already knows the contents of the file, which is why we do not pass the
        text to the formatting function.
        """
        # Ensure that the LSP has the latest version of the file contents
        try:
            changes = self.lsp.format(path=self.path)
        except Exception:  # noqa: S110
            pass
        else:
            for change in changes or []:
                if new := change.get("newText"):
                    # Convert line/char to position ranges
                    range_ = change.get("range", {})
                    start = range_.get("start", {})
                    start_line = start.get("line", 0)
                    start_char = start.get("character", 0)
                    end = range_.get("end", {})
                    end_line = end.get("line", 0)
                    end_char = end.get("character", 0)
                    segment = range_to_slice(
                        start_line, start_char, end_line, end_char, text
                    )
                    text = f"{text[: segment.start]}{new}{text[segment.stop :]}"
            text = text.rstrip()

        return text

    def __repr__(self) -> str:
        """Return representation of the formatter as a string."""
        return f"{self.lsp.name.title()}Formatter()"



"""Local Python interpreter kernel implementation."""

from __future__ import annotations

import ast
import code
import getpass
import logging
import sys
import threading
import traceback
from asyncio import to_thread
from functools import update_wrapper
from linecache import cache as line_cache
from typing import TYPE_CHECKING, cast

from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import Python3TracebackLexer

from euporie.core.app.current import get_app
from euporie.core.kernel.base import BaseKernel, KernelInfo, MsgCallbacks

if TYPE_CHECKING:
    from typing import Any, Callable, Unpack

    from euporie.core.tabs.kernel import KernelTab

log = logging.getLogger(__name__)


class LocalPythonKernel(BaseKernel):
    """Run code in a local Python interpreter."""

    @classmethod
    def variants(cls) -> list[KernelInfo]:
        """Return available kernel specifications."""
        return [
            KernelInfo(
                name="local-python",
                display_name="Local Python",
                factory=cls,
                kind="new",
                type=cls,
            )
        ]

    # Thread-local storage for callbacks during execution
    _thread_local = threading.local()

    def __init__(
        self,
        kernel_tab: KernelTab,
        default_callbacks: MsgCallbacks | None = None,
        allow_stdin: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize the local Python interpreter kernel.

        Args:
            kernel_tab: The notebook this kernel belongs to
            allow_stdin: Whether the kernel is allowed to request input
            default_callbacks: The default callbacks to use on receipt of a message
            connection_file: Not used for local kernel
            **kwargs: Additional keyword arguments
        """
        super().__init__(
            kernel_tab=kernel_tab,
            allow_stdin=allow_stdin,
            default_callbacks=default_callbacks,
        )
        # Create interpreter with callback for error handling
        self.locals: dict[str, object] = {}
        self.execution_count = 0
        self.kc = None
        self._input_event = threading.Event()
        self._input_buffer: str | None = None
        self.hook_manager = HookManager(self)

    @property
    def missing(self) -> bool:
        """Return whether the kernel is missing."""
        return False

    @property
    def id(self) -> str | None:
        """Return the kernel ID."""
        return "local-python"

    def info(
        self,
        set_kernel_info: Callable[[dict[str, Any]], None] | None = None,
        set_status: Callable[[str], None] | None = None,
    ) -> None:
        """Request information about the kernel.

        Args:
            set_kernel_info: Callback to set kernel info
            set_status: Callback to set kernel status
        """
        kernel_info = {
            "language_info": {
                "name": "python",
                "version": sys.version.split()[0],
                "mimetype": "text/x-python",
                "file_extension": ".py",
                "pygments_lexer": "python3",
                "codemirror_mode": {
                    "name": "python",
                    "version": sys.version_info.major,
                },
                "nbconvert_exporter": "python",
            },
            "implementation": sys.implementation.name,
            "implementation_version": sys.version.split()[0],
            "banner": f"Python {sys.version}",
            "help_links": ["https://euporie.readthedocs.io/"],
            "status": "ok",
        }

        if callable(set_kernel_info):
            set_kernel_info(kernel_info)

        if callable(set_status):
            set_status(self.status)

    async def start_async(self) -> None:
        """Start the local interpreter."""
        self.error = None
        self.status = "idle"
        self.locals.clear()
        self.execution_count = 0
        if callable(
            set_execution_count := self.default_callbacks.get("set_execution_count")
        ):
            set_execution_count(self.execution_count)

    @property
    def spec(self) -> dict[str, str]:
        """The kernelspec metadata for the current kernel instance."""
        return {
            "name": "local-python",
            "display_name": "Local Python",
            "language": "python",
        }

    def stop(self, cb: Callable | None = None, wait: bool = False) -> None:
        """Stop the local interpreter."""
        self.status = "stopped"
        if callable(cb):
            cb()

    def showtraceback(
        self, filename: str, cb: Callable[[dict[str, Any], bool], None] | None
    ) -> None:
        """Format and display tracebacks for exceptions."""
        typ, value, tb = sys.exc_info()
        stack_summary = list(traceback.extract_tb(tb))

        # Filter items from stack prior to executed code
        for i, frame in enumerate(stack_summary):
            if frame.filename == filename:
                stack_summary = stack_summary[i:]
                break
        else:
            stack_summary = []

        # Format the traceback text
        traceback_text = "".join(
            [
                *traceback.format_list(stack_summary),
                *traceback.format_exception_only(typ, value),
            ]
        )

        # Color the traceback using Pygments
        colored_traceback = highlight(
            traceback_text,
            Python3TracebackLexer(),
            Terminal256Formatter(style=get_app().config.syntax_theme),
        ).rstrip()

        # Send the error through the callback
        if callable(cb):
            cb(
                {
                    "output_type": "error",
                    "ename": "Exception" if typ is None else typ.__name__,
                    "evalue": str(value),
                    "traceback": colored_traceback.splitlines(),
                },
                True,
            )

    def _execute_code(
        self,
        body: list[ast.stmt],
        last: ast.stmt | None,
        filename: str,
        callbacks: MsgCallbacks,
    ) -> None:
        """Execute code in the interpreter.

        Args:
            body: List of statements to execute
            last: Optional final expression to evaluate
            filename: Source filename for tracebacks
            callbacks: Message callbacks
        """
        # Store callbacks in thread local storage
        self._thread_local.callbacks = callbacks
        add_output = callbacks["add_output"]

        with self.hook_manager:
            # Execute body
            try:
                if body:
                    exec(  # noqa: S102
                        compile(
                            ast.Module(body=body, type_ignores=[]),
                            filename=filename,
                            mode="exec",
                        ),
                        self.locals,
                    )
                # Last statement is an expression - eval it
                if last is not None:
                    exec(  # noqa: S102
                        compile(
                            ast.Interactive([last]),
                            filename=filename,
                            mode="single",
                        ),
                        self.locals,
                    )
            except SystemExit:
                get_app().exit()
            except Exception:
                self.showtraceback(filename, cb=add_output)
                if callable(done := callbacks.get("done")):
                    done({"status": "error"})

    async def run_async(
        self, source: str, **local_callbacks: Unpack[MsgCallbacks]
    ) -> None:
        """Execute code in the local interpreter."""
        callbacks = MsgCallbacks(
            {
                **self.default_callbacks,
                **cast(
                    "MsgCallbacks",
                    {k: v for k, v in local_callbacks.items() if v is not None},
                ),
            }
        )

        self.status = "busy"

        # Set execution count
        self.execution_count += 1
        if callable(set_execution_count := callbacks.get("set_execution_count")):
            set_execution_count(self.execution_count)

        # Add source to line cache (for display in tracebacks)
        filename = f"<input_{self.execution_count}>"
        line_cache[filename] = (
            len(source),
            None,
            source.splitlines(keepends=True),
            filename,
        )
        try:
            # Parse the source into an AST
            tree = ast.parse(source, filename=filename)
        except Exception:
            # Check for syntax errors
            self.showtraceback(filename, cb=callbacks["add_output"])
            return
        else:
            if not tree.body:
                return

        # Split into statements and final expression
        body = tree.body
        last = None
        if isinstance(tree.body[-1], ast.Expr):
            last = tree.body.pop()

        # Execute the code in a thread
        await to_thread(self._execute_code, body, last, filename, callbacks)

        self.status = "idle"

        if callable(done := callbacks.get("done")):
            done({"status": "ok"})

    async def complete_async(self, source: str, cursor_pos: int) -> list[dict]:
        """Get code completions."""
        import rlcompleter

        completer = rlcompleter.Completer(self.locals)

        # Find the last word before cursor
        tokens = source[:cursor_pos].split()
        if not tokens:
            return []

        word = tokens[-1]
        completions = []

        # Get all possible completions
        i = 0
        while True:
            completion = completer.complete(word, i)
            if completion is None:
                break
            completions.append({"text": completion, "start_position": -len(word)})
            i += 1

        return completions

    async def inspect_async(
        self,
        source: str,
        cursor_pos: int,
        detail_level: int = 0,
        timeout: int = 2,
    ) -> dict[str, Any]:
        """Get code inspection/documentation."""
        import inspect

        if not source:
            return {}

        # Find the start of the word (going backwards from cursor)
        start = cursor_pos
        while (start >= 0 and source[start - 1].isalnum()) or source[start - 1] in "._":
            start -= 1
        # Find the end of the word (going forwards from cursor)
        end = cursor_pos
        while end < len(source) and (source[end].isalnum() or source[end] in "._"):
            end += 1
        # Extract the complete word
        obj_name = source[start:end].strip()
        if not obj_name:
            return {}
        try:
            obj = eval(obj_name, self.locals)  # noqa: S307
        except Exception:
            return {}
        else:
            if doc := inspect.getdoc(obj):
                return {"text/plain": doc}
        return {}

    async def is_complete_async(
        self,
        source: str,
        timeout: int | float = 0.1,
    ) -> dict[str, Any]:
        """Check if code is complete."""
        try:
            compiled = code.compile_command(
                source, f"<input_{self.execution_count}>", "exec"
            )
        except Exception:
            status = "invalid"
        else:
            status = "incomplete" if compiled is None else "complete"

        result = {"status": status, "indent": "    " if source[-1:] in ":({[" else ""}
        return result

    def input(self, text: str) -> None:
        """Send input to the kernel.

        Args:
            text: The input text to provide
        """
        self._input_buffer = text
        self._input_event.set()

    def interrupt(self) -> None:
        """Interrupt the kernel."""
        log.warning("Cannot interrupt kernel %r", self)

    async def restart_async(
        self, wait: bool = False, cb: Callable | None = None
    ) -> None:
        """Restart the kernel."""
        await self.start_async()
        if callable(cb):
            cb({"status": "ok"})

    async def shutdown_async(self, wait: bool = False) -> None:
        """Shutdown the kernel."""
        self.stop()


def get_display_data(obj: Any) -> tuple[dict[str, Any], dict[str, Any]]:
    """Get display data and metadata for an object.

    Args:
        obj: Object to get display data for

    Returns:
        Tuple of (data, metadata) dictionaries
    """
    data = {}
    metadata = {}

    if hasattr(obj, "_repr_mimebundle_"):
        output = obj._repr_mimebundle_()
        if isinstance(output, tuple):
            data, metadata = output
        else:
            data = output
    else:
        for method, mime in [
            ("_repr_latex_", "text/latex"),
            ("_repr_html_", "text/html"),
            ("_repr_markdown_", "text/markdown"),
            ("_repr_svg_", "image/svg+xml"),
            ("_repr_jpeg_", "image/jpeg"),
            ("_repr_png_", "image/png"),
        ]:
            if hasattr(obj, method) and (output := getattr(obj, method)()) is not None:
                data[mime] = output
        if not data:
            data = {"text/plain": repr(obj)}

    return data, metadata


class BaseHook:
    """Base class providing access to thread-specific callbacks."""

    def __init__(self, kernel: LocalPythonKernel) -> None:
        """Initialize the base hook.

        Args:
            kernel: The kernel instance to hook
        """
        self._kernel = kernel

    @property
    def callbacks(self) -> MsgCallbacks:
        """Get callbacks for current thread."""
        return getattr(
            self._kernel._thread_local, "callbacks", self._kernel.default_callbacks
        )


class DisplayHook(BaseHook):
    """Hook for sys.displayhook that dispatches to thread-specific callbacks."""

    def __call__(self, value: Any) -> None:
        """Handle display of values."""
        if value is None:
            return

        if callbacks := self.callbacks:
            # Store value in kernel locals
            self._kernel.locals[f"_{self._kernel.execution_count}"] = value

            # Get display data and metadata
            data, metadata = get_display_data(value)

            if callable(callback := callbacks.get("add_output")):
                callback(
                    {
                        "output_type": "execute_result",
                        "execution_count": self._kernel.execution_count,
                        "data": data,
                        "metadata": metadata,
                    },
                    True,
                )


class DisplayGlobal(BaseHook):
    """A display() function that dispatches to thread-specific callbacks.

    This class implements the global display() function used to show rich output
    in notebooks. It routes display calls to the appropriate output callbacks.
    """

    def __call__(
        self,
        *objs: Any,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        metadata: dict[str, Any] | None = None,
        transient: dict[str, Any] | None = None,
        display_id: str | None = None,
        raw: bool = False,
        clear: bool = False,
        **kwargs: Any,
    ) -> None:
        """Handle display of values in the notebook.

        Args:
            *objs: Objects to display. Each object will be displayed in sequence.
            include: List of MIME types to include in the output. If specified,
                only these MIME types will be included.
            exclude: List of MIME types to exclude from the output. These MIME
                types will be filtered out.
            metadata: Additional metadata to attach to the display data.
            transient: Transient data that is used for display but not persisted
                in the notebook document.
            display_id: Unique identifier for the display. Can be used to update
                this display output later.
            raw: If True, skip MIME type transformation/formatting of the objects.
            clear: If True, clear the output before displaying new content.
            **kwargs: Additional display arguments passed to the frontend.
        """
        add_output = self.callbacks.get("add_output")
        if not callable(add_output):
            return

        for obj in objs:
            data, obj_metadata = get_display_data(obj)

            # Filter mime types
            if include:
                data = {k: v for k, v in data.items() if k in include}
            if exclude:
                data = {k: v for k, v in data.items() if k not in exclude}

            # Merge metadata
            if metadata:
                obj_metadata.update(metadata)

            add_output(
                {
                    "output_type": "display_data",
                    "data": data,
                    "metadata": obj_metadata,
                    "transient": transient,
                    "display_id": display_id,
                },
                True,
            )


class InputBuiltin(BaseHook):
    """Hook for input and getpass that dispatches to thread-specific callbacks."""

    def __init__(self, kernel: LocalPythonKernel, is_password: bool = False) -> None:
        """Initialize the input hook.

        Args:
            kernel: The kernel instance to hook
            is_password: Whether this is for password input
        """
        super().__init__(kernel)
        self._is_password = is_password
        update_wrapper(self, input)

    def __call__(self, prompt: str = "", stream: Any = None) -> str:
        """Get input from user via callback."""
        if (callbacks := self.callbacks) and (get_input := callbacks.get("get_input")):
            # Clear any previous input
            self._kernel._input_event.clear()
            self._kernel._input_buffer = None

            # Request input via callback
            get_input(prompt, self._is_password)

            # Wait for input to be provided
            self._kernel._input_event.wait()

            # Return the input, or empty string if none provided
            return self._kernel._input_buffer or ""
        return ""


class StreamWrapper(BaseHook):
    """Hook for stdout/stderr that dispatches to thread-specific callbacks."""

    def __init__(self, name: str, kernel: LocalPythonKernel) -> None:
        """Initialize the stream hook.

        Args:
            name: Name of the stream (stdout/stderr)
            kernel: The kernel instance to hook
        """
        BaseHook.__init__(self, kernel)
        self.name = name
        self._thread_local = threading.local()

    @property
    def buffer(self) -> str:
        """Get the thread-local buffer for this stream."""
        if not hasattr(self._thread_local, "buffer"):
            self._thread_local.buffer = ""
        return self._thread_local.buffer

    @buffer.setter
    def buffer(self, value: str) -> None:
        self._thread_local.buffer = value

    def _send(self, text: str, callbacks: MsgCallbacks) -> None:
        """Send text to the frontend via callback."""
        if callable(callback := callbacks.get("add_output")):
            callback(
                {"output_type": "stream", "name": self.name, "text": text},
                True,
            )

    def write(self, text: str) -> int:
        """Write output using callback for current thread."""
        if not isinstance(text, str):
            raise TypeError(f"write() argument must be str, not {type(text)}")
        if not text or not (callbacks := self.callbacks):
            return 0

        # Handle any buffered content plus new text
        all_text = self.buffer + text
        lines = all_text.splitlines(keepends=True)
        if lines[-1].endswith("\n"):
            self.buffer = ""
        else:
            self.buffer = lines[-1]
            lines = lines[:-1]
        if lines:
            # Send complete lines immediately
            self._send("".join(lines), callbacks)

        return len(text)

    def flush(self) -> None:
        """Flush any buffered content."""
        if self.buffer and (callbacks := self.callbacks):
            self._send(self.buffer, callbacks)
            self.buffer = ""


class HookManager:
    """Context manager for hooking stdout/stderr/displayhook."""

    def __init__(self, kernel: LocalPythonKernel) -> None:
        """Initialize the hook manager.

        Args:
            kernel: The kernel instance to hook
        """
        # Create hook instances
        self.stdout = StreamWrapper("stdout", kernel)
        self.stderr = StreamWrapper("stderr", kernel)
        self.displayhook = DisplayHook(kernel)
        self.display = DisplayGlobal(kernel)
        self.input = InputBuiltin(kernel, is_password=False)
        self.getpass = InputBuiltin(kernel, is_password=True)
        # Store original objects
        self.og_stdout = sys.stdout
        self.og_stderr = sys.stderr
        self.og_displayhook = sys.displayhook
        self.og_getpass = getpass.getpass
        # Track hook depth
        self._depth = 0
        self._kernel = kernel

    def __enter__(self) -> None:
        """Replace objects with hooks."""
        if self._depth == 0:
            # Replace system streams
            sys.stdout = self.stdout
            sys.stderr = self.stderr
            sys.displayhook = self.displayhook
            # Replace getpass
            getpass.getpass = self.getpass
            # Add input to kernel locals
            self._kernel.locals["input"] = self.input
            # Add display to kernel locals
            self._kernel.locals["display"] = self.display
        self._depth += 1

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Restore original objects."""
        self._depth -= 1
        if self._depth == 0:
            # Restore system streams
            sys.stdout = self.og_stdout
            sys.stderr = self.og_stderr
            sys.displayhook = self.og_displayhook
            # Restore getpass
            getpass.getpass = self.og_getpass
            # Remove input from kernel locals
            self._kernel.locals.pop("input", None)
            self._kernel.locals.pop("display", None)
        # Flush any remaining stream outputs
        self.stdout.flush()
        self.stderr.flush()



"""Defines various tool bars displayed at the bottom of apps."""

# Define toolbar buffer names

SEARCH_BAR_BUFFER = "SEARCH_BAR_BUFFER"
COMMAND_BAR_BUFFER = "COMMAND_BAR_BUFFER"

BAR_BUFFERS = {
    SEARCH_BAR_BUFFER,
    COMMAND_BAR_BUFFER,
}



"""Define settings for the notebook app tabs."""

from euporie.core.config import add_setting

add_setting(
    name="show_scroll_bar",
    group="euporie.notebook.tabs.notebook",
    title="scroll bar",
    flags=["--show-scroll-bar"],
    type_=bool,
    help_="Show the scroll bar",
    default=True,
    description="""
        Whether the scroll bar should be shown on the right of the screen.
    """,
)

add_setting(
    name="show_side_bar",
    group="euporie.notebook.widgets.side_bar",
    flags=["--show-side-bar"],
    type_=bool,
    title="side-bar",
    help_="Show the side-bar",
    default=False,
    schema={
        "type": "boolean",
    },
    description="""
        Whether the side-bar should be shown at the side of the screen.
    """,
)



"""Define a file browser widget."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.key_binding.key_bindings import KeyBindings, KeyBindingsBase
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
)
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.layout.screen import WritePosition
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core.app.current import get_app
from euporie.core.border import InsetGrid
from euporie.core.layout.containers import HSplit, VSplit, Window
from euporie.core.layout.decor import FocusedStyle
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.widgets.decor import Border
from euporie.core.widgets.forms import Button, Text

if TYPE_CHECKING:
    from typing import Callable

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from upath.core import PT

    from euporie.core.bars.status import StatusBarFields

log = logging.getLogger(__name__)


FILE_ICONS = {
    # default
    "file": "",
    "dir": "",
    # file extensions (vim-devicons)
    ".styl": "",
    ".sass": "",
    ".scss": "",
    ".htm": "",
    ".html": "",
    ".slim": "",
    ".haml": "",
    ".ejs": "",
    ".css": "",
    ".less": "",
    ".md": "",
    ".mdx": "",
    ".markdown": "",
    ".rmd": "",
    ".json": "",
    ".webmanifest": "",
    ".js": "",
    ".mjs": "",
    ".jsx": "",
    ".rb": "",
    ".gemspec": "",
    ".rake": "",
    ".php": "",
    ".py": "",
    ".pyc": "",
    ".pyo": "",
    ".pyd": "",
    ".coffee": "",
    ".mustache": "",
    ".hbs": "",
    ".conf": "",
    ".ini": "",
    ".yml": "",
    ".yaml": "",
    ".toml": "",
    ".bat": "",
    ".mk": "",
    ".twig": "",
    ".cpp": "",
    ".c++": "",
    ".cxx": "",
    ".cc": "",
    ".cp": "",
    ".c": "",
    ".cs": "",
    ".h": "",
    ".hh": "",
    ".hpp": "",
    ".hxx": "",
    ".hs": "",
    ".lhs": "",
    ".nix": "",
    ".lua": "",
    ".java": "",
    ".sh": "",
    ".fish": "",
    ".bash": "",
    ".zsh": "",
    ".ksh": "",
    ".csh": "",
    ".awk": "",
    ".ps1": "",
    ".ml": "λ",
    ".mli": "λ",
    ".diff": "",
    ".db": "",
    ".sql": "",
    ".dump": "",
    ".clj": "",
    ".cljc": "",
    ".cljs": "",
    ".edn": "",
    ".scala": "",
    ".go": "",
    ".dart": "",
    ".xul": "",
    ".sln": "",
    ".suo": "",
    ".pl": "",
    ".pm": "",
    ".t": "",
    ".rss": "",
    ".f#": "",
    ".fsscript": "",
    ".fsx": "",
    ".fs": "",
    ".fsi": "",
    ".rs": "",
    ".rlib": "",
    ".d": "",
    ".erl": "",
    ".hrl": "",
    ".ex": "",
    ".exs": "",
    ".eex": "",
    ".leex": "",
    ".heex": "",
    ".vim": "",
    ".ai": "",
    ".psd": "",
    ".psb": "",
    ".ts": "",
    ".tsx": "",
    ".jl": "",
    ".pp": "",
    ".vue": "﵂",
    ".elm": "",
    ".swift": "",
    ".xcplayground": "",
    ".tex": "ﭨ",
    ".r": "ﳒ",
    ".rproj": "鉶",
    ".sol": "ﲹ",
    ".pem": "",
    # file names (vim-devicons) (case-insensitive not supported in lf)
    "gruntfile.coffee": "",
    "gruntfile.js": "",
    "gruntfile.ls": "",
    "gulpfile.coffee": "",
    "gulpfile.js": "",
    "gulpfile.ls": "",
    "mix.lock": "",
    "dropbox": "",
    ".ds_store": "",
    ".gitconfig": "",
    ".gitignore": "",
    ".gitattributes": "",
    ".gitlab-ci.yml": "",
    ".bashrc": "",
    ".zshrc": "",
    ".zshenv": "",
    ".zprofile": "",
    ".vimrc": "",
    ".gvimrc": "",
    "_vimrc": "",
    "_gvimrc": "",
    ".bashprofile": "",
    "favicon.ico": "",
    "license": "",
    "node_modules": "",
    "react.jsx": "",
    "procfile": "",
    "dockerfile": "",
    "docker-compose.yml": "",
    "rakefile": "",
    "config.ru": "",
    "gemfile": "",
    "makefile": "",
    "cmakelists.txt": "",
    "robots.txt": "ﮧ",
    # file names (case-sensitive adaptations)
    "Gruntfile.coffee": "",
    "Gruntfile.js": "",
    "Gruntfile.ls": "",
    "Gulpfile.coffee": "",
    "Gulpfile.js": "",
    "Gulpfile.ls": "",
    "Dropbox": "",
    ".DS_Store": "",
    "LICENSE": "",
    "React.jsx": "",
    "Procfile": "",
    "Dockerfile": "",
    "Docker-compose.yml": "",
    "Rakefile": "",
    "Gemfile": "",
    "Makefile": "",
    "CMakeLists.txt": "",
    # file patterns (file name adaptations)
    "jquery.min.js": "",
    "angular.min.js": "",
    "backbone.min.js": "",
    "require.min.js": "",
    "materialize.min.js": "",
    "materialize.min.css": "",
    "mootools.min.js": "",
    "vimrc": "",
    "Vagrantfile": "",
    # archives or compressed (extensions from dircolors defaults)
    ".tar": "",
    ".tgz": "",
    ".arc": "",
    ".arj": "",
    ".taz": "",
    ".lha": "",
    ".lz4": "",
    ".lzh": "",
    ".lzma": "",
    ".tlz": "",
    ".txz": "",
    ".tzo": "",
    ".t7z": "",
    ".zip": "",
    ".z": "",
    ".dz": "",
    ".gz": "",
    ".lrz": "",
    ".lz": "",
    ".lzo": "",
    ".xz": "",
    ".zst": "",
    ".tzst": "",
    ".bz2": "",
    ".bz": "",
    ".tbz": "",
    ".tbz2": "",
    ".tz": "",
    ".deb": "",
    ".rpm": "",
    ".jar": "",
    ".war": "",
    ".ear": "",
    ".sar": "",
    ".rar": "",
    ".alz": "",
    ".ace": "",
    ".zoo": "",
    ".cpio": "",
    ".7z": "",
    ".rz": "",
    ".cab": "",
    ".wim": "",
    ".swm": "",
    ".dwm": "",
    ".esd": "",
    # image formats (extensions from dircolors defaults)
    ".jpg": "",
    ".jpeg": "",
    ".mjpg": "",
    ".mjpeg": "",
    ".webp": "",
    ".ico": "",
    ".gif": "",
    ".bmp": "",
    ".pbm": "",
    ".pgm": "",
    ".ppm": "",
    ".tga": "",
    ".xbm": "",
    ".xpm": "",
    ".tif": "",
    ".tiff": "",
    ".png": "",
    ".svg": "",
    ".svgz": "",
    ".mng": "",
    ".pcx": "",
    ".mov": "",
    ".mpg": "",
    ".mpeg": "",
    ".m2v": "",
    ".mkv": "",
    ".webm": "",
    ".ogm": "",
    ".mp4": "",
    ".m4v": "",
    ".mp4v": "",
    ".vob": "",
    ".qt": "",
    ".nuv": "",
    ".wmv": "",
    ".asf": "",
    ".rm": "",
    ".rmvb": "",
    ".flc": "",
    ".avi": "",
    ".fli": "",
    ".flv": "",
    ".gl": "",
    ".dl": "",
    ".xcf": "",
    ".xwd": "",
    ".yuv": "",
    ".cgm": "",
    ".emf": "",
    ".ogv": "",
    ".ogx": "",
    # audio formats (extensions from dircolors defaults)
    ".aac": "",
    ".au": "",
    ".flac": "",
    ".m4a": "",
    ".mid": "",
    ".midi": "",
    ".mka": "",
    ".mp3": "",
    ".mpc": "",
    ".ogg": "",
    ".ra": "",
    ".wav": "",
    ".oga": "",
    ".opus": "",
    ".spx": "",
    ".xspf": "",
    # documents
    ".pdf": "",
    ".doc": "",
    ".docx": "",
    ".ipynb": "ﴬ",
}


def is_dir(path: str | Path) -> bool | None:
    """Check if a path is a directory."""
    test_path = UPath(path)
    try:
        return test_path.is_dir()
    except (ValueError, PermissionError, TypeError):
        return None


class FileBrowserControl(UIControl):
    """A control for browsing a filesystem."""

    def __init__(
        self,
        path: Path | None = None,
        on_chdir: Callable[[FileBrowserControl], None] | None = None,
        on_select: Callable[[FileBrowserControl], None] | None = None,
        on_open: Callable[[FileBrowserControl], None] | None = None,
        window: Window | None = None,
    ) -> None:
        """Initialize a new file browser instance."""
        self.dir = path or UPath(".")
        self.hovered: int | None = None
        self.selected: int | None = None
        self._dir_cache: FastDictCache[tuple[Path], list[tuple[bool, Path]]] = (
            FastDictCache(get_value=self.load_path, size=1)
        )
        self.on_select = Event(self, on_select)
        self.on_chdir = Event(self, on_chdir)
        self.on_open = Event(self, on_open)

        self.window = window

        self.on_chdir.fire()

        self.key_bindings = kb = KeyBindings()

        @kb.add("up")
        @kb.add("<scroll-up>")
        def _move_up(event: KeyPressEvent) -> None:
            self.move_cursor_up()

        @kb.add("down")
        @kb.add("<scroll-down>")
        def _move_down(event: KeyPressEvent) -> None:
            self.move_cursor_down()

        @kb.add("home")
        def _home(event: KeyPressEvent) -> None:
            self.select(0)

        @kb.add("end")
        def _end(event: KeyPressEvent) -> None:
            self.select(len(self.contents) - 1)

        @kb.add("left")
        def _up(event: KeyPressEvent) -> None:
            self.dir = self.dir.parent

        @kb.add("space")
        @kb.add("enter")
        @kb.add("right")
        def _open(event: KeyPressEvent) -> None:
            return self.open_path()

    @property
    def contents(self) -> list[tuple[bool, Path]]:
        """Return the contents of the current folder."""
        return self._dir_cache[(self.dir,)]

    @property
    def dir(self) -> Path:
        """Return the current folder path."""
        return self._dir

    @dir.setter
    def dir(self, value: PT) -> None:
        """Set the current folder path."""
        dir_path = UPath(value)
        try:
            dir_path = dir_path.resolve()
        except NotImplementedError:
            pass
        if is_dir(dir_path):
            self._dir = dir_path
        else:
            log.warning("'%s' is not a directory, not changing directory", value)

    @property
    def path(self) -> Path:
        """Return the current selected path."""
        return self.contents[self.selected or 0][1]

    @staticmethod
    def load_path(path: Path) -> list[tuple[bool, Path]]:
        """Return the contents of a folder."""
        paths = [] if path.parent == path else [path / ".."]
        try:
            paths += list(path.iterdir())
        except PermissionError:
            pass
        is_dirs = []
        for child in paths:
            child_is_dir = is_dir(child)
            if child_is_dir is None:
                child_is_dir = True
            is_dirs.append(child_is_dir)
        return sorted(zip(is_dirs, paths), key=lambda x: (not x[0], x[1].name))

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the content for this user control."""
        paths = self.contents

        def get_line(i: int) -> StyleAndTextTuples:
            if i > len(paths) - 1:
                return []
            is_dir, child = paths[i]

            row = child.name

            if get_app().config.show_file_icons:
                icon = (
                    FILE_ICONS["dir"]
                    if is_dir
                    else FILE_ICONS.get(child.suffix)
                    or FILE_ICONS.get(child.name)
                    or FILE_ICONS["file"]
                )
                row = f"{icon} {row}"

            style = "class:row"
            if i % 2:
                style += " class:alt-row"
            if i == self.hovered:
                style += " class:hovered"
            if i == self.selected:
                style += " class:selection"

            return [(style, f" {row} ".ljust(width))]

        return UIContent(
            get_line=get_line,
            line_count=len(paths),
            cursor_position=Point(0, self.selected or 0),
            show_cursor=False,
        )

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events."""
        row = mouse_event.position.y
        app = get_app()
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_DOWN
        ):
            app.layout.current_control = self
            app.mouse_limits = None
            self.hovered = None
            return self.select(row, open_file=True)
        elif mouse_event.event_type == MouseEventType.MOUSE_MOVE:
            # Mark item as hovered if mouse is over the control
            if (
                self.window is not None
                and (info := self.window.render_info) is not None
            ):
                rowcol_to_yx = info._rowcol_to_yx
                abs_mouse_pos = (
                    mouse_event.position.x + info._x_offset,
                    mouse_event.position.y + info._y_offset - info.vertical_scroll,
                )
                if abs_mouse_pos == app.mouse_position:
                    row_col_vals = rowcol_to_yx.values()
                    y_min, x_min = min(row_col_vals)
                    y_max, x_max = max(row_col_vals)
                    app.mouse_limits = WritePosition(
                        xpos=x_min,
                        ypos=y_min,
                        width=x_max - x_min + 1,
                        height=y_max - y_min + 1,
                    )
                    return self.hover(row)
                else:
                    # Clear mouse limits if mouse is outside control
                    app.mouse_limits = None
                    self.hovered = None
                    return None

        return NotImplemented

    def select(self, row: int | None, open_file: bool = False) -> NotImplementedOrNone:
        """Select a file in the browser."""
        if row is None:
            row = 0
        row = min(max(0, row), len(self.contents) - 1)
        if self.selected != row:
            self.selected = row
            self.on_select.fire()
        elif open_file:
            self.open_path()
        return None

    def hover(self, row: int | None) -> NotImplementedOrNone:
        """Hover a file in the browser."""
        if row is not None:
            row = min(max(0, row), len(self.contents) - 1)
            if self.hovered != row:
                self.hovered = row
                return None
        return NotImplemented

    def open_path(self) -> None:
        """Open the selected file."""
        if self.selected is not None:
            is_dir, path = self.contents[self.selected]
            if is_dir:
                self.dir = path.resolve()
                self.hover(self.hovered)
                self.selected = None
                self.on_chdir.fire()
            else:
                self.on_open.fire()

    def move_cursor_down(self) -> None:
        """Request to move the cursor down."""
        index = self.selected
        if index is None:
            index = 0
        else:
            index += 1
        self.select(index)

    def move_cursor_up(self) -> None:
        """Request to move the cursor up."""
        index = self.selected
        if index is None:
            index = len(self.contents)
        else:
            index -= 1
        self.select(index)

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Key bindings specific to this user control."""
        return self.key_bindings

    def is_focusable(self) -> bool:
        """Determine that the file_browser is focusable."""
        return True

    def __pt_status__(self) -> StatusBarFields:
        """Show the selected or hovered path in the statusbar."""
        if self.hovered is not None:
            return [[("", str(self.contents[self.hovered][1]))]], []
        elif self.selected is not None:
            return [[("", str(self.contents[self.selected][1]))]], []
        return [], []


class FileBrowser:
    """A file browser."""

    completer = PathCompleter(only_directories=True)

    def __init__(
        self,
        path: Path | None = None,
        on_select: Callable[[Path], None] | None = None,
        on_open: Callable[[Path], None] | None = None,
        on_chdir: Callable[[Path], None] | None = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
        style: str = "",
        show_address_bar: FilterOrBool = True,
    ) -> None:
        """Create a new instance."""

        def _accept_path(buffer: Buffer) -> bool:
            control.dir = Path(buffer.text)
            return True

        def _validate_path(path: str) -> bool:
            return is_dir(path) or False

        text = Text(
            validation=_validate_path,
            accept_handler=_accept_path,
            completer=self.completer,
        )
        self.control = control = FileBrowserControl(
            path=path,
            on_chdir=lambda x: setattr(text, "text", str(x.dir)),
        )
        if on_select is not None:
            control.on_select += (
                lambda x: on_select(x.path) if callable(on_select) else None
            )
        if on_chdir is not None:
            control.on_chdir += (
                lambda x: on_chdir(x.path) if callable(on_chdir) else None
            )
        if on_open is not None:
            control.on_open += lambda x: on_open(x.path) if callable(on_open) else None

        self.container = HSplit(
            [
                ConditionalContainer(
                    VSplit(
                        [
                            FocusedStyle(text),
                            FocusedStyle(
                                Button(
                                    "Go",
                                    on_click=lambda x: setattr(
                                        control, "dir", text.text
                                    ),
                                )
                            ),
                        ]
                    ),
                    filter=to_filter(show_address_bar),
                ),
                FocusedStyle(
                    Border(
                        VSplit(
                            [
                                window := Window(
                                    control,
                                    style="class:face",
                                ),
                                MarginContainer(ScrollbarMargin(), target=window),
                            ]
                        ),
                        border=InsetGrid,
                        style="class:input,inset,border",
                    ),
                    style_hover="",
                ),
            ],
            style="class:file-browser " + style,
            width=width,
            height=height,
        )
        # Set control's window so it can determine it's position for mouse-over
        control.window = window

    def __pt_container__(self) -> AnyContainer:
        """Return the tree-view container's content."""
        return self.container

    def __pt_status__(self) -> StatusBarFields:
        """Show the selected or hovered path in the statusbar."""
        return self.control.__pt_status__()



"""Defines tab settings."""

from prompt_toolkit.filters import buffer_has_focus

from euporie.core.config import add_setting

add_setting(
    name="backup_on_save",
    group="euporie.core.tabs.base",
    flags=["--backup-on-save"],
    type_=bool,
    help_="Create backups before saving files",
    default=False,
    description="""
        Determines whether a backup file should be created before saving a file.

        If set to ``True``, the original file will be copied to a new file with the
        same name but prefixed with ``.`` and suffixed with ``.bak`` before writing the
        updated contents.
    """,
)

add_setting(
    name="kernel_name",
    group="euporie.core.tabs.kernel",
    flags=["--kernel-name", "--kernel"],
    type_=str,
    help_="The name of the kernel to start by default",
    default="python3",
    description="""
        The name of the kernel selected automatically by the console app or in new
        notebooks. If set to an empty string, the user will be asked which kernel
        to launch.
    """,
)

add_setting(
    name="record_cell_timing",
    group="euporie.core.tabs.kernel",
    title="cell timing recording",
    flags=["--record-cell-timing"],
    type_=bool,
    help_="Should timing data be recorded in cell metadata.",
    default=False,
    schema={
        "type": "boolean",
    },
    description="""
        When set, execution timing data will be recorded in cell metadata.
    """,
)

add_setting(
    name="show_remote_inputs",
    group="euporie.core.tabs.kernel",
    flags=["--show-remote-inputs"],
    type_=bool,
    help_="Display inputs sent to the kernel by other clients",
    default=True,
    description="""
        If set to `True`, all code input sent to the kernel by any client will be
        displayed.

        If set to `False`, only inputs sent to the kernel by the current instance
        of euporie will be displayed, and all other inputs will be ignored.

    """,
)

add_setting(
    name="show_remote_outputs",
    group="euporie.core.tabs.kernel",
    flags=["--show-remote-outputs"],
    type_=bool,
    help_="Display kernel outputs triggered by other clients",
    default=True,
    description="""
        If set to `False`, only outputs generated by code input from the current
        instance of euporie will be displayed, and all other outputs will be
        ignored.

        If set to `True`, all outputs generated by the kernel will be
        displayed.
    """,
)

# euporie.core.tabs.notebook

add_setting(
    name="save_widget_state",
    group="euporie.core.tabs.notebook",
    flags=["--save-widget-state"],
    type_=bool,
    help_="Save a notebook's widget state in the notebook metadata",
    default=True,
    description="""
        When set to ``True``, the state of any widgets in the current notebook will
        be saves in the notebook's metadata. This enables widgets to be displayed
        when the notebook is re-opened without having to re-run the notebook.
    """,
)

add_setting(
    name="max_notebook_width",
    group="euporie.core.tabs.notebook",
    flags=["--max-notebook-width"],
    type_=int,
    help_="Maximum width of notebooks",
    default=120,
    schema={
        "minimum": 1,
    },
    description="""
        The maximum width at which to display a notebook.
    """,
)

add_setting(
    name="expand",
    group="euporie.core.tabs.notebook",
    flags=["--expand"],
    type_=bool,
    help_="Use the full width to display notebooks",
    default=False,
    description="""
        Whether the notebook page should expand to fill the available width
    """,
    cmd_filter=~buffer_has_focus,
)



"""Contain main format conversion function."""

from __future__ import annotations

import logging
import mimetypes
from functools import lru_cache
from typing import TYPE_CHECKING

from upath import UPath
from upath._stat import UPathStatResult
from upath.implementations.http import HTTPPath

if TYPE_CHECKING:
    from pathlib import Path


log = logging.getLogger(__name__)


MIME_FORMATS = {
    "image/svg+xml": "svg",
    "image/png": "png",
    "image/jpeg": "jpeg",
    "image/gif": "gif",
    "application/pdf": "pdf",
    "text/html": "html",
    "text/latex": "latex",
    "application/x-latex": "latex",
    "text/markdown": "markdown",
    "text/x-markdown": "markdown",
    "text/*": "ansi",
    "stream/std*": "ansi",
    "*": "ansi",
}


@lru_cache
def get_mime(path: Path | str) -> str | None:
    """Attempt to determine the mime-type of a path."""
    if isinstance(path, str):
        path = UPath(path)
    try:
        path = path.resolve()
    except Exception:
        log.debug("Cannot resolve '%s'", path)

    mime = None

    # Read from path of data URI
    if path.exists() and isinstance(stat := path.stat(), UPathStatResult):
        mime = stat.as_info().get("mimetype")

    # If we have a web-address, ensure we have a url
    # Check http-headers and ensure we have a url
    if not mime and isinstance(path, HTTPPath) and path._url is not None:
        from fsspec.asyn import sync

        # Get parsed url
        url = path._url.geturl()
        # Get the fsspec fs
        fs = path.fs
        # Ensure we have a session
        session = sync(fs.loop, fs.set_session)
        # Use HEAD requests if the server allows it, falling back to GETs
        for method in (session.head, session.get):
            r = sync(fs.loop, method, url, allow_redirects=True)
            try:
                r.raise_for_status()
            except Exception:
                log.debug("Request failed: %s", r)
                continue
            else:
                content_type = r.headers.get("Content-Type")
                if content_type is not None:
                    mime = content_type.partition(";")[0]
                    break
    # Try using magic
    if not mime:
        try:
            import magic
        except ModuleNotFoundError:
            pass
        else:
            try:
                with path.open(mode="rb") as f:
                    mime = magic.from_buffer(f.read(2048), mime=True)
            except FileNotFoundError:
                pass

    # Guess from file-extension
    if not mime and path.suffix:
        # Check for Jupyter notebooks by extension
        if path.suffix == ".ipynb":
            return "application/x-ipynb+json"
        else:
            mime, _ = mimetypes.guess_type(path)

    return mime


@lru_cache
def get_format(path: Path | str, default: str = "") -> str:
    """Attempt to guess the format of a path."""
    if isinstance(path, str):
        path = UPath(path)
    if not default:
        default = "html" if isinstance(path, HTTPPath) else "ansi"
    mime = get_mime(path)
    return MIME_FORMATS.get(mime, default) if mime else default



"""Style related functions."""

from __future__ import annotations

import logging
from colorsys import hls_to_rgb, rgb_to_hls
from functools import cache, partial
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.styles.defaults import default_ui_style
from prompt_toolkit.styles.style import Style
from pygments.styles import get_style_by_name as pyg_get_style_by_name

if TYPE_CHECKING:
    from typing import Any

    from pygments.style import Style as PygmentsStyle


log = logging.getLogger(__name__)


KERNEL_STATUS_REPR = {
    "stopped": "⨂",
    "starting": "◍",
    "idle": "○",
    "busy": "●",
    "error": "☹",
}


DEFAULT_COLORS = {
    "bg": "#232627",
    "fg": "#fcfcfc",
    "ansiblack": "#000000",
    "ansired": "#cc0403",
    "ansigreen": "#19cb00",
    "ansiyellow": "#cecb00",
    "ansiblue": "#0d73cc",
    "ansipurple": "#9841bb",
    "ansimagenta": "#cb1ed1",
    "ansicyan": "#0dcdcd",
    "ansiwhite": "#dddddd",
    "ansibrightblack": "#767676",
    "ansigray": "#767676",
    "ansibrightred": "#f2201f",
    "ansibrightgreen": "#23fd00",
    "ansibrightyellow": "#fffd00",
    "ansibrightblue": "#1a8fff",
    "ansibrightpurple": "#fd28ff",
    "ansibrightmagenta": "#fd28ff",
    "ansibrightcyan": "#14ffff",
    "ansibrightwhite": "#ffffff",
}


MIME_STYLE = [
    ("mime-stream-stderr", "fg:ansired"),
]

HTML_STYLE = [
    ("html h1", "bold underline"),
    ("html h1 border", "fg:ansiyellow nounderline"),
    ("html h2", "bold"),
    ("html h2 border", "fg:grey nobold"),
    ("html h3", "bold"),
    ("html h4", "bold italic"),
    ("html h5", "underline"),
    ("html h6", "italic"),
    # ("html code", "bg:#333"),
    # ("html kbd", "fg:#fff bg:#333"),
    # ("html samp", "fg:#fff bg:#333"),
    # ("html b", "bold"),
    # ("html strong", "bold"),
    # ("html i", "italic"),
    # ("html em", "italic"),
    # ("html cite", "italic"),
    # ("html dfn", "italic"),
    # ("html var", "italic"),
    # ("html u", "underline"),
    # ("html ins", "underline"),
    # ("html s", "strike"),
    # ("html del", "strike"),
    # ("html mark", "fg:black bg:ansiyellow"),
    # ("html hr", "fg:ansired"),
    ("html ul bullet", "fg:ansiyellow"),
    ("html ol bullet", "fg:ansicyan"),
    ("html blockquote", "fg:ansipurple"),
    ("html blockquote margin", "fg:grey"),
    ("html th", "bold"),
    ("html img", "bg:cyan fg:black"),
    ("html img border", "fg:cyan bg:default"),
    ("html caption", "italic"),
]


LOG_STYLE = [
    ("log.level.nonset", "fg:ansigray"),
    ("log.level.debug", "fg:ansigreen"),
    ("log.level.info", "fg:ansiblue"),
    ("log.level.warning", "fg:ansiyellow"),
    ("log.level.error", "fg:ansired"),
    ("log.level.critical", "fg:ansiwhite bg:ansired bold"),
]

IPYWIDGET_STYLE = [
    ("ipywidget", "bg:default"),
    ("ipywidget focused", ""),
    ("ipywidget button face", "fg:black bg:#d4d0c8"),
    ("ipywidget button face selection", "bg:#e9e7e3"),
    ("ipywidget button face success", "bg:ansigreen"),
    ("ipywidget button face info", "bg:ansicyan"),
    ("ipywidget button face warning", "bg:ansiyellow"),
    ("ipywidget button face danger", "bg:ansired"),
    ("ipywidget button border right", "fg:#606060"),
    ("ipywidget button border bottom", "fg:#606060"),
    ("ipywidget button border left", "fg:#ffffff"),
    ("ipywidget button border top", "fg:#ffffff"),
    ("ipywidget button face disabled", "fg:#888888"),
    ("ipywidget selection border right", "fg:#ffffff"),
    ("ipywidget selection border bottom", "fg:#ffffff"),
    ("ipywidget selection border left", "fg:#606060"),
    ("ipywidget selection border top", "fg:#606060"),
    # ("ipywidget success border left", "fg:ansibrightgreen"),
    # ("ipywidget success border top", "fg:ansibrightgreen"),
    # ("ipywidget info border left", "fg:ansibrightcyan"),
    # ("ipywidget info border top", "fg:ansibrightcyan"),
    # ("ipywidget warning border left", "fg:ansibrightyellow"),
    # ("ipywidget warning border top", "fg:ansibrightyellow"),
    # ("ipywidget danger border left", "fg:ansibrightred"),
    # ("ipywidget danger border top", "fg:ansibrightred"),
    # ("ipywidget border left selection", "fg:#606060"),
    # ("ipywidget border top selection", "fg:#606060"),
    # ("ipywidget success border right selection", "fg:ansibrightgreen"),
    # ("ipywidget success border bottom selection", "fg:ansibrightgreen"),
    # ("ipywidget success border right selection", "fg:ansibrightgreen"),
    # ("ipywidget success border bottom selection", "fg:ansibrightgreen"),
    # ("ipywidget info border right selection", "fg:ansibrightcyan"),
    # ("ipywidget info border bottom selection", "fg:ansibrightcyan"),
    # ("ipywidget danger border right selection", "fg:ansibrightred"),
    # ("ipywidget danger border bottom selection", "fg:ansibrightred"),
    ("ipywidget text text-area", "fg:black bg:white"),
    ("ipywidget text text-area disabled", "fg:#888888"),
    ("ipywidget text text-area focused", "fg:black bg:white"),
    ("ipywidget text placeholder", "fg:#AAAAAA bg:white"),
    ("ipywidget text border right", "fg:#E9E7E3"),
    ("ipywidget text border top", "fg:#606060"),
    ("ipywidget text border bottom", "fg:#E9E7E3"),
    ("ipywidget text border left", "fg:#606060"),
    ("ipywidget text border top invalid", "fg:ansidarkred"),
    ("ipywidget text border right invalid", "fg:ansired"),
    ("ipywidget text border bottom invalid", "fg:ansired"),
    ("ipywidget text border left invalid", "fg:ansidarkred"),
    ("ipywidget slider track selection", "fg:ansiblue"),
    ("ipywidget slider handle focused", "fg:white"),
    ("ipywidget slider handle selection", ""),
    ("ipywidget slider handle selection focused", "fg:ansiblue"),
    ("ipywidget progress", "fg:ansidarkblue bg:#d4d0c8"),
    ("ipywidget progress success", "fg:ansigreen"),
    ("ipywidget progress info", "fg:ansicyan"),
    ("ipywidget progress warning", "fg:ansiyellow"),
    ("ipywidget progress danger", "fg:ansired"),
    ("ipywidget progress border", "bg:default"),
    ("ipywidget progress border top", "fg:#E9E7E3"),
    ("ipywidget progress border right", "fg:#606060"),
    ("ipywidget progress border bottom", "fg:#606060"),
    ("ipywidget progress border left", "fg:#E9E7E3"),
    ("ipywidget dropdown dropdown.menu", "fg:#000000 bg:#ffffff"),
    ("ipywidget dropdown dropdown.menu hovered", "fg:#ffffff bg:ansiblue"),
    ("ipywidget checkbox selection", "fg:default"),
    ("ipywidget checkbox prefix selection", "fg:ansiblue"),
    ("ipywidget checkbox disabled", "fg:#888888"),
    ("ipywidget checkbox prefix selection disabled", "fg:#888888"),
    ("ipywidget valid prefix", "fg:ansired"),
    ("ipywidget valid prefix selection", "fg:ansigreen"),
    ("ipywidget radio-buttons selection", "fg:default"),
    ("ipywidget radio-buttons selection disabled", "fg:#888888"),
    ("ipywidget radio-buttons prefix selection", "fg:ansiblue"),
    ("ipywidget select face", "fg:black bg:white"),
    ("ipywidget select face selection", "fg:white bg:ansiblue"),
    ("ipywidget select face hovered", "fg:black bg:#f0f0f0"),
    ("ipywidget select face hovered selection", "fg:white bg:ansiblue"),
    ("ipywidget select border top", "fg:#606060"),
    ("ipywidget select border right", "fg:#E9E7E3"),
    ("ipywidget select border bottom", "fg:#E9E7E3"),
    ("ipywidget select border left", "fg:#606060"),
    ("ipywidget select disabled", "fg:#888888"),
    ("ipywidget select selection disabled", "fg:#888888"),
    ("ipywidget inset border right", "fg:#E9E7E3"),
    ("ipywidget inset border top", "fg:#606060"),
    ("ipywidget inset border bottom", "fg:#E9E7E3"),
    ("ipywidget inset border left", "fg:#606060"),
    ("ipywidget swatch border right", "fg:#E9E7E3"),
    ("ipywidget swatch border top", "fg:#606060"),
    ("ipywidget swatch border bottom", "fg:#E9E7E3"),
    ("ipywidget swatch border left", "fg:#606060"),
    ("ipywidget tabbed-split tab-bar tab active border top default", "fg:ansiblue"),
    ("ipywidget tabbed-split tab-bar tab active border top success", "fg:ansigreen"),
    ("ipywidget tabbed-split tab-bar tab active border top info", "fg:ansicyan"),
    ("ipywidget tabbed-split tab-bar tab active border top warning", "fg:ansiyellow"),
    ("ipywidget tabbed-split tab-bar tab active border top danger", "fg:ansired"),
    ("ipywidget accordion border success", "fg:ansigreen"),
    ("ipywidget accordion border info", "fg:ansicyan"),
    ("ipywidget accordion border warning", "fg:ansiyellow"),
    ("ipywidget accordion border danger", "fg:ansired"),
    ("ipywidget accordion selection", "fg:ansiblue"),
]


class ColorPaletteColor:
    """A representation of a color with adjustment methods."""

    _cache: SimpleCache[tuple[str, float, float, float, bool], ColorPaletteColor] = (
        SimpleCache()
    )

    def __init__(self, base: str, _base_override: str = "") -> None:
        """Create a new color.

        Args:
            base: The base color as a hexadecimal string.
            _base_override: An optional base color override.
        """
        self.base_hex = DEFAULT_COLORS.get(base, base)
        self.base = _base_override or base

        color = self.base_hex.lstrip("#")
        self.red, self.green, self.blue = (
            int(color[0:2], 16) / 255,
            int(color[2:4], 16) / 255,
            int(color[4:6], 16) / 255,
        )

        self.hue, self.brightness, self.saturation = rgb_to_hls(
            self.red, self.green, self.blue
        )

        self.is_light = self.brightness > 0.5

    def _adjust_abs(
        self, hue: float = 0.0, brightness: float = 0.0, saturation: float = 0.0
    ) -> ColorPaletteColor:
        hue = (self.hue + hue) % 1
        brightness = max(min(1, self.brightness + brightness), 0)
        saturation = max(min(1, self.saturation + saturation), 0)

        r, g, b = hls_to_rgb(hue, brightness, saturation)
        new_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        return ColorPaletteColor(new_color)

    def _adjust_rel(
        self, hue: float = 0.0, brightness: float = 0.0, saturation: float = 0.0
    ) -> ColorPaletteColor:
        hue = min(max(0, hue), 1)
        brightness = min(max(-1, brightness), 1)
        saturation = min(max(-1, saturation), 1)

        new_hue = self.hue + (self.hue * (hue < 0) + (1 - self.hue) * (hue > 0)) * hue

        new_brightness = (
            self.brightness
            + (
                self.brightness * (brightness < 0)
                + (1 - self.brightness) * (brightness > 0)
            )
            * brightness
        )

        new_saturation = (
            self.saturation
            + (
                self.saturation * (saturation < 0)
                + (1 - self.saturation) * (saturation > 0)
            )
            * saturation
        )

        r, g, b = hls_to_rgb(new_hue, new_brightness, new_saturation)
        new_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        return ColorPaletteColor(new_color)

    def _adjust(
        self,
        hue: float = 0.0,
        brightness: float = 0.0,
        saturation: float = 0.0,
        rel: bool = True,
    ) -> ColorPaletteColor:
        """Perform a relative of absolute color adjustment.

        Args:
            hue: The hue adjustment.
            brightness: The brightness adjustment.
            saturation: The saturation adjustment.
            rel: If True, perform a relative adjustment.

        Returns:
            ColorPaletteColor: The adjusted color.
        """
        if rel:
            return self._adjust_rel(hue, brightness, saturation)
        else:
            return self._adjust_abs(hue, brightness, saturation)

    def adjust(
        self,
        hue: float = 0.0,
        brightness: float = 0.0,
        saturation: float = 0.0,
        rel: bool = True,
    ) -> ColorPaletteColor:
        """Adjust the hue, saturation, or brightness of the color.

        Args:
            hue: The hue adjustment.
            brightness: The brightness adjustment.
            saturation: The saturation adjustment.
            rel: If True, perform a relative adjustment.

        Returns:
            ColorPaletteColor: The adjusted color.
        """
        key = (self.base_hex, hue, brightness, saturation, rel)
        return self._cache.get(
            key, partial(self._adjust, hue, brightness, saturation, rel)
        )

    def lighter(self, amount: float, rel: bool = True) -> ColorPaletteColor:
        """Make the color lighter.

        Args:
            amount: The amount to lighten the color by.
            rel: If True, perform a relative adjustment.

        Returns:
            ColorPaletteColor: The lighter color.
        """
        return self.adjust(brightness=amount, rel=rel)

    def darker(self, amount: float, rel: bool = True) -> ColorPaletteColor:
        """Make the color darker.

        Args:
            amount: The amount to darken the color by.
            rel: If True, perform a relative adjustment.

        Returns:
            ColorPaletteColor: The darker color.
        """
        return self.adjust(brightness=-amount, rel=rel)

    def more(self, amount: float, rel: bool = True) -> ColorPaletteColor:
        """Make bright colors darker and dark colors brighter.

        Args:
            amount: The amount to adjust the color by.
            rel: If True, perform a relative adjustment.

        Returns:
            ColorPaletteColor: The adjusted color.
        """
        if self.is_light:
            amount *= -1
        return self.adjust(brightness=amount, rel=rel)

    def less(self, amount: float, rel: bool = True) -> ColorPaletteColor:
        """Make bright colors brighter and dark colors darker.

        Args:
            amount: The amount to adjust the color by.
            rel: If True, perform a relative adjustment.

        Returns:
            ColorPaletteColor: The adjusted color.
        """
        if self.is_light:
            amount *= -1
        return self.adjust(brightness=-amount, rel=rel)

    def towards(self, other: ColorPaletteColor, amount: float) -> ColorPaletteColor:
        """Interpolate between two colors."""
        amount = min(max(0, amount), 1)
        r = (other.red - self.red) * amount + self.red
        g = (other.green - self.green) * amount + self.green
        b = (other.blue - self.blue) * amount + self.blue
        new_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        return ColorPaletteColor(new_color)

    def __repr__(self) -> str:
        """Return a representation of the color."""
        return f"Color({self.base})"

    def __str__(self) -> str:
        """Return a string representation of the color."""
        return self.base_hex


class ColorPalette:
    """Define a collection of colors."""

    def __init__(self) -> None:
        """Create a new color-palette."""
        self.colors: dict[str, ColorPaletteColor] = {}

    def add_color(self, name: str, base: str, _base_override: str = "") -> ColorPalette:
        """Add a color to the palette."""
        self.colors[name] = ColorPaletteColor(base, _base_override)
        return self

    def __getattr__(self, name: str) -> Any:
        """Enable access of palette colors via dotted attributes.

        Args:
            name: The name of the attribute to access.

        Returns:
            The color-palette color.

        """
        return self.colors[name]


def build_style(
    cp: ColorPalette,
    have_term_colors: bool = True,
) -> Style:
    """Create an application style based on the given color palette."""
    style_dict = {
        # The default style is merged at this point so full styles can be
        # overridden. For example, this allows us to switch off the underline
        # status of cursor-line.
        **dict(default_ui_style().style_rules),
        "default": f"fg:{cp.bg.base} bg:{cp.bg.base}",
        # Remove non-breaking space style from PTK
        "nbsp": "nounderline fg:default",
        # Logo
        "logo": "fg:#dd0000",
        # Pattern
        "pattern": f"fg:{cp.bg.more(0.05)}",
        # Chrome
        "chrome": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "tab-padding": f"fg:{cp.bg.more(0.2)} bg:{cp.bg.base}",
        # Statusbar
        # "status": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.less(0.15)}",
        "status": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "status-field": f"bg:{cp.fg.more(0.1)} fg:{cp.bg.more(0.1)} reverse",
        "status-sep": f"bg:{cp.bg.more(0.05)} fg:{cp.bg.more(0.1)} reverse",
        # Menus & Menu bar
        "menu": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "menu bar": f"bg:{cp.bg.less(0.15)}",
        "menu disabled": f"fg:{cp.fg.more(0.05).towards(cp.bg, 0.75)}",
        "menu shortcut": f"fg:{cp.fg.more(0.4)}",
        "menu shortcut disabled": f"fg:{cp.fg.more(0.4).towards(cp.bg, 0.5)}",
        "menu prefix": f"fg:{cp.fg.more(0.2)}",
        "menu prefix disabled": f"fg:{cp.fg.more(0.2).towards(cp.bg, 0.5)}",
        "menu selection": f"bg:{cp.hl.more(1)} fg:{cp.hl} reverse",
        "menu selection shortcut": f"bg:{cp.hl.more(1).more(0.05)} fg:{cp.hl} reverse",
        "menu selection prefix": f"bg:{cp.hl.more(1).more(0.05)} fg:{cp.hl} reverse",
        "menu border": f"fg:{cp.bg.more(0.15)} bg:{cp.bg.more(0.05)}",
        "menu border selection": f"fg:{cp.bg.more(0.15)} bg:{cp.hl} noreverse",
        # Tab bar
        "app-tab-bar": f"bg:{cp.bg.less(0.15)}",
        "app-tab-bar border": f"fg:{cp.bg.more(0.1)}",
        "app-tab-bar tab inactive": f"fg:{cp.fg.more(0.5)}",
        "app-tab-bar tab inactive border": f"fg:{cp.bg.more(0.15)}",
        "app-tab-bar tab active": "bold fg:default bg:default",
        "app-tab-bar tab active close": "fg:darkred",
        "app-tab-bar tab active border top": f"fg:{cp.hl} bg:{cp.bg.less(0.15)}",
        # Tabs
        "loading": "fg:#888888",
        # Buffer
        "line-number": f"bg:{cp.fg.more(0.5)} fg:{cp.bg.more(0.05)} reverse",
        "line-number.current": f"bg:orange fg:{cp.bg.more(0.1)} bold",
        "line-number edge": f"bg:{cp.bg.darker(0.1)}",
        "line-number.current edge": f"bg:{cp.bg.darker(0.1)}",
        "cursor-line": f"bg:{cp.bg.more(0.05)}",
        "cursor-line search": f"bg:{cp.bg.more(0.02)}",
        "cursor-line search.current": f"bg:{cp.bg.more(0.02)}",
        "cursor-line incsearch": "bg:ansibrightyellow",
        "cursor-line incsearch.current": "bg:ansibrightgreen",
        "matching-bracket.cursor": "fg:yellow bold",
        "matching-bracket.other": "fg:yellow bold",
        "trailing-whitespace": f"fg:{cp.fg.more(0.66)}",
        "tab": f"fg:{cp.fg.more(0.66)}",
        # Search results
        "search": f"bg:{cp.bg.more(0.05)}",
        "search.current": f"bg:{cp.bg.more(0.05)}",
        "incsearch": "bg:ansibrightyellow",
        "incsearch.current": "bg:ansibrightgreen",
        # Inputs
        "kernel-input": f"fg:default bg:{cp.bg.more(0.02)}",
        # Cells
        # "cell cell.selection": f"bg:{cp.bg.towards(cp.hl, 0.05)}",
        # "cell edit": f"bg:{cp.bg.towards(cp.hl.adjust(hue=-0.3333, rel=False), 0.025)}",
        "cell border": f"fg:{cp.bg.more(0.25)}",
        "cell border cell.selection": f"fg:{cp.hl.more(0.2)}",
        "cell border edit": f"fg:{cp.hl.adjust(hue=-0.3333, rel=False)}",
        "cell input prompt": "fg:blue",
        "cell output prompt": "fg:red",
        "cell show outputs": f"bg:{cp.fg.more(0.5)} fg:{cp.bg.more(0.05)} reverse",
        "cell show inputs": f"bg:{cp.fg.more(0.5)} fg:{cp.bg.more(0.05)} reverse",
        "cell show inputs border": f"bg:{cp.bg.darker(0.1)} fg:{cp.bg.more(0.05)} reverse",
        "cell show outputs border": f"bg:{cp.bg.darker(0.1)} fg:{cp.bg.more(0.05)} reverse",
        # Scrollbars
        "scrollbar": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.15)}",
        "scrollbar.background": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.15)}",
        "scrollbar.arrow": f"bg:{cp.bg.more(0.75)} fg:{cp.bg.more(0.20)} reverse",
        "scrollbar.start": "",
        # "scrollbar.start": f"fg:{cp.bg.more(0.75)} bg:{cp.bg.more(0.25)}",
        "scrollbar.button": f"bg:{cp.bg.more(0.75)} fg:{cp.bg.more(0.75)} reverse",
        "scrollbar.end": f"bg:{cp.bg.more(0.15)} fg:{cp.bg.more(0.75)} reverse",
        # Overflow margin
        "overflow": f"fg:{cp.fg.more(0.5)}",
        # Dialogs
        "dialog dialog-title": f"bg:white fg:{cp.hl.darker(0.25)} bold reverse",
        "dialog": f"fg:{cp.fg.base} bg:{cp.bg.darker(0.1)}",
        "dialog text-area": f"bg:{cp.bg.lighter(0.05)}",
        "dialog input text text-area": f"fg:default bg:{cp.bg.less(0.1)}",
        "dialog text-area last-line": "nounderline",
        "dialog border": f"fg:{cp.bg.darker(0.1).more(0.1)}",
        # Horizontals rule
        "hr": "fg:ansired",
        # Toolbars
        "toolbar": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "toolbar.title": f"fg:{cp.fg.more(0.1)} bg:{cp.bg.more(0.1)}",
        # Search bar
        "search-toolbar": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        # Command bar
        "toolbar menu": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "toolbar menu completion": f"fg:{cp.fg.more(0.1)} bg:{cp.bg.more(0.1)}",
        "toolbar menu completion current": f"fg:{cp.hl} bg:{cp.fg} reverse",
        "toolbar menu overflow": f"fg:{cp.fg.more(0.5)}",
        "toolbar menu meta": f"bg:{cp.bg.more(0.25)} bold",
        # Completions menu
        "menu completion-keyword": "fg:#d700af",
        "menu completion-function": "fg:#005faf",
        "menu completion-class": "fg:#008700",
        "menu completion-statement": "fg:#5f0000",
        "menu completion-instance": "fg:#d75f00",
        "menu completion-module": "fg:#d70000",
        "menu completion-magic": "fg:#9841bb",
        "menu completion-path": "fg:#aa8800",
        "menu completion-dict-key": "fg:#ddbb00",
        "menu selection completion-keyword": (
            f"bg:{ColorPaletteColor('#d700af').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-function": (
            f"bg:{ColorPaletteColor('#005faf').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-class": (
            f"bg:{ColorPaletteColor('#008700').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-statement": (
            f"bg:{ColorPaletteColor('#5f0000').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-instance": (
            f"bg:{ColorPaletteColor('#d75f00').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-module": (
            f"bg:{ColorPaletteColor('#d70000').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-magic": (
            f"bg:{ColorPaletteColor('#888888').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        "menu selection completion-path": (
            f"bg:{ColorPaletteColor('#aa8800').lighter(0.75)} fg:{cp.hl} reverse"
        ),
        # Log
        "log.level.nonset": "fg:grey",
        "log.level.debug": "fg:green",
        "log.level.info": "fg:blue",
        "log.level.warning": "fg:yellow",
        "log.level.error": "fg:red",
        "log.level.critical": "fg:red bold",
        "log.ref": "fg:grey",
        "log.date": "fg:#00875f",
        # File browser
        "file-browser border": f"fg:{cp.bg.more(0.5)}",
        "file-browser face": f"bg:{cp.bg.lighter(0.1)}",
        "file-browser face row alt-row": f"bg:{cp.bg.lighter(0.1).more(0.01)}",
        "file-browser face row hovered": f"bg:{cp.bg.more(0.2)}",
        "file-browser face row selection": f"bg:{cp.fg} fg:{cp.hl} reverse",
        # Shortcuts
        "shortcuts.group": f"bg:{cp.bg.more(0.4)} bold underline",
        # "shortcuts.row": f"bg:{cp.bg.base} nobold",
        "shortcuts.row alt": f"bg:{cp.bg.more(0.1)}",
        "shortcuts.row key": "bold",
        # Palette
        "palette.item": f"fg:{cp.fg.more(0.05)} bg:{cp.bg.more(0.05)}",
        "palette.item.alt": f"bg:{cp.bg.more(0.15)}",
        "palette.item.selected": f"bg:{cp.hl.more(1)} fg:{cp.hl} reverse",
        # Pager
        "pager": f"bg:{cp.bg.more(0.05)}",
        "pager.border": f"fg:{cp.bg.towards(cp.ansiblack, 0.15)} reverse",
        # Markdown
        "markdown code": f"bg:{cp.bg.more(0.15)}",
        "markdown code block": f"bg:{cp.bg.less(0.2)}",
        "markdown code block border": f"fg:{cp.bg.more(0.25)}",
        "markdown table border": f"fg:{cp.bg.more(0.75)}",
        # Drop-shadow
        "drop-shadow inner": f"fg:{cp.bg.towards(cp.ansiblack, 0.3)}",
        "drop-shadow outer": f"fg:{cp.bg.towards(cp.ansiblack, 0.2)} bg:{cp.bg.towards(cp.ansiblack, 0.05)}",
        # Side-bar
        "side_bar": f"bg:{cp.bg.less(0.15)}",
        "side_bar title": f"fg:{cp.hl}",
        "side_bar title text": f"fg:default bg:{cp.bg.less(0.15).more(0.01)}",
        "side_bar border": f"fg:{cp.bg.towards(cp.ansiblack, 0.3)}",
        "side_bar border outer": f"bg:{cp.bg}",
        "side_bar buttons": f"bg:{cp.bg.less(0.15)}",
        "side_bar buttons hovered": f"fg:{cp.hl}",
        "side_bar buttons separator": f"bg:{cp.bg.less(0.15)} fg:{cp.bg.less(0.15)}",
        "side_bar buttons selection": f"bg:{cp.fg} fg:{cp.hl} reverse",
        "side_bar buttons separator selection before": (
            f"bg:{cp.bg.less(0.15)} fg:{cp.hl} reverse"
        ),
        "side_bar buttons separator selection after": (
            f"fg:{cp.hl} bg:{cp.bg.less(0.15)} noreverse"
        ),
        # Tabbed split
        "tabbed-split border": f"fg:{cp.bg.more(0.2)}",
        "tabbed-split border left": f"bg:{cp.bg.more(0.025)}",
        "tabbed-split border right": f"bg:{cp.bg.more(0.025)}",
        "tabbed-split border bottom left": f"bg:{cp.bg}",
        "tabbed-split border bottom right": f"bg:{cp.bg}",
        "tabbed-split page": f"bg:{cp.bg.more(0.025)}",
        "dialog tabbed-split border bottom right": f"bg:{cp.bg.darker(0.1)}",
        "dialog tabbed-split border bottom left": f"bg:{cp.bg.darker(0.1)}",
        "tabbed-split tab-bar tab inactive": f"fg:{cp.bg.more(0.3)}",
        "tabbed-split tab-bar tab inactive title": f"bg:{cp.bg.darker(0.05)}",
        "tabbed-split tab-bar tab inactive border left": f"bg:{cp.bg.darker(0.05)}",
        "tabbed-split tab-bar tab inactive border right": f"bg:{cp.bg.darker(0.05)}",
        "tabbed-split tab-bar tab active": f"bold fg:{cp.fg}",
        "tabbed-split tab-bar tab active title": f"bg:{cp.bg.more(0.025)}",
        "tabbed-split tab-bar tab active close": "fg:darkred",
        # Ipywidgets
        "ipywidget focused": f"bg:{cp.bg.more(0.05)}",
        "ipywidget slider track": f"fg:{cp.fg.darker(0.5)}",
        "ipywidget slider arrow": f"fg:{cp.fg.darker(0.25)}",
        "ipywidget slider handle": f"fg:{cp.fg.darker(0.25)}",
        "ipywidget accordion border default": f"fg:{cp.bg.more(0.2)}",
        ## Styled borders
        "success border top": "fg:ansibrightgreen",
        "success border left": "fg:ansibrightgreen",
        "success border bottom": f"fg:{cp.ansigreen.darker(0.5)}",
        "success border right": f"fg:{cp.ansigreen.darker(0.5)}",
        "info border top": "fg:ansibrightcyan",
        "info border left": "fg:ansibrightcyan",
        "info border bottom": f"fg:{cp.ansicyan.darker(0.5)}",
        "info border right": f"fg:{cp.ansicyan.darker(0.5)}",
        "warning border top": "fg:ansibrightyellow",
        "warning border left": "fg:ansibrightyellow",
        "warning border bottom": f"fg:{cp.ansiyellow.darker(0.5)}",
        "warning border right": f"fg:{cp.ansiyellow.darker(0.5)}",
        "danger border top": "fg:ansibrightred",
        "danger border left": "fg:ansibrightred",
        "danger border bottom": f"fg:{cp.ansired.darker(0.5)}",
        "danger border right": f"fg:{cp.ansired.darker(0.5)}",
        ## Selected styled borders
        "selection success border top": f"fg:{cp.ansigreen.darker(0.5)}",
        "selection success border left": f"fg:{cp.ansigreen.darker(0.5)}",
        "selection success border bottom": "fg:ansibrightgreen",
        "selection success border right": "fg:ansibrightgreen",
        "selection info border left": f"fg:{cp.ansicyan.darker(0.5)}",
        "selection info border top": f"fg:{cp.ansicyan.darker(0.5)}",
        "selection info border right": "fg:ansibrightcyan",
        "selection info border bottom": "fg:ansibrightcyan",
        "selection warning border top": f"fg:{cp.ansiyellow.darker(0.5)}",
        "selection warning border left": f"fg:{cp.ansiyellow.darker(0.5)}",
        "selection warning border bottom": "fg:ansibrightyellow",
        "selection warning border right": "fg:ansibrightyellow",
        "selection danger border top": f"fg:{cp.ansired.darker(0.5)}",
        "selection danger border left": f"fg:{cp.ansired.darker(0.5)}",
        "selection danger border bottom": "fg:ansibrightred",
        "selection danger border right": "fg:ansibrightred",
        # Selected faces
        "selection success face": f"bg:{cp.ansigreen.darker(0.05)}",
        "selection info face": f"bg:{cp.ansicyan.darker(0.05)}",
        "selection warning face": f"bg:{cp.ansiyellow.darker(0.05)}",
        "selection danger face": f"bg:{cp.ansired.darker(0.05)}",
        # Hovered faces
        "input hovered face": f"fg:default bg:{cp.bg.more(0.2)}",
        "focused hovered success face": f"bg:{cp.ansigreen.lighter(0.05)}",
        "focused hovered info face": f"bg:{cp.ansicyan.lighter(0.05)}",
        "focused hovered warning face": f"bg:{cp.ansiyellow.lighter(0.05)}",
        "focused hovered danger face": f"bg:{cp.ansired.lighter(0.05)}",
        # Text areas
        "text-area focused": "noreverse",
        "text-area selected": "noreverse",
        "text-area focused selected": "reverse",
        # Buttons
        "input button face": f"fg:default bg:{cp.bg.more(0.05)}",
        "input button face hovered": f"fg:{cp.fg} bg:{cp.bg.more(0.2)}",
        "input button face selection": f"bg:{cp.fg} fg:{cp.bg.more(0.05)} reverse",
        "input button face focused": f"bg:{cp.fg} fg:{cp.bg.towards(cp.hl, 0.1)} reverse",
        # Input widgets
        "input border top": f"fg:{cp.bg.lighter(0.5)}",
        "input border left": f"fg:{cp.bg.lighter(0.5)}",
        "input border bottom": f"fg:{cp.bg.darker(0.25)}",
        "input border right": f"fg:{cp.bg.darker(0.25)}",
        "input border top focused": f"fg:{cp.hl.lighter(0.5)}",
        "input border left focused": f"fg:{cp.hl.lighter(0.5)}",
        "input border bottom focused": f"fg:{cp.hl.darker(0.5)}",
        "input border right focused": f"fg:{cp.hl.darker(0.5)}",
        "input border top selection": f"fg:{cp.bg.darker(0.5)}",
        "input border left selection": f"fg:{cp.bg.darker(0.5)}",
        "input border bottom selection": f"fg:{cp.bg.lighter(0.5)}",
        "input border right selection": f"fg:{cp.bg.lighter(0.5)}",
        "input border top selection focused": f"fg:{cp.hl.darker(0.5)}",
        "input border left selection focused": f"fg:{cp.hl.darker(0.5)}",
        "input border bottom selection focused": f"fg:{cp.hl.lighter(0.5)}",
        "input border right selection focused": f"fg:{cp.hl.lighter(0.5)}",
        "input inset border bottom": f"fg:{cp.bg.lighter(0.5)}",
        "input inset border right": f"fg:{cp.bg.lighter(0.5)}",
        "input inset border top": f"fg:{cp.bg.darker(0.25)}",
        "input inset border left": f"fg:{cp.bg.darker(0.25)}",
        "input inset border bottom focused": f"fg:{cp.hl.lighter(0.5)}",
        "input inset border right focused": f"fg:{cp.hl.lighter(0.5)}",
        "input inset border top focused": f"fg:{cp.hl.darker(0.5)}",
        "input inset border left focused": f"fg:{cp.hl.darker(0.5)}",
        "input inset border bottom selection": f"fg:{cp.bg.darker(0.5)}",
        "input inset border right selection": f"fg:{cp.bg.darker(0.5)}",
        "input inset border top selection": f"fg:{cp.bg.lighter(0.5)}",
        "input inset border left selection": f"fg:{cp.bg.lighter(0.5)}",
        "input inset border bottom selection focused": f"fg:{cp.hl.darker(0.5)}",
        "input inset border right selection focused": f"fg:{cp.hl.darker(0.5)}",
        "input inset border top selection focused": f"fg:{cp.hl.lighter(0.5)}",
        "input inset border left selection focused": f"fg:{cp.hl.lighter(0.5)}",
        "input text placeholder": f"fg:{cp.fg.more(0.6)}",
        "input text text-area": f"fg:default bg:{cp.bg.lighter(0.1)}",
        "input text border top": f"fg:{cp.bg.darker(0.5)}",
        "input text border right": f"fg:{cp.bg.lighter(0.25)}",
        "input text border bottom": f"fg:{cp.bg.lighter(0.25)}",
        "input text border left": f"fg:{cp.bg.darker(0.5)}",
        "input text border top focused": f"fg:{cp.hl.darker(0.5)}",
        "input text border right focused": f"fg:{cp.hl.lighter(0.5)}",
        "input text border bottom focused": f"fg:{cp.hl.lighter(0.5)}",
        "input text border left focused": f"fg:{cp.hl.darker(0.5)}",
        "input text border top invalid": "fg:ansidarkred",
        "input text border right invalid": "fg:ansired",
        "input text border bottom invalid": "fg:ansired",
        "input text border left invalid": "fg:ansidarkred",
        "input radio-buttons prefix selection focused": f"fg:{cp.hl}",
        "input slider arrow": f"fg:{cp.fg.darker(0.25)}",
        "input slider track": f"fg:{cp.fg.darker(0.5)}",
        "input slider track selection": f"fg:{cp.hl}",
        "input slider handle": f"fg:{cp.fg.darker(0.25)}",
        "input slider handle focused": f"fg:{cp.fg}",
        "input slider handle selection focused": f"fg:{cp.hl}",
        "input dropdown dropdown.menu": f"bg:{cp.bg.more(0.05)}",
        "input dropdown dropdown.menu hovered": f"bg:{cp.hl}",
        "input select face": f"bg:{cp.bg.lighter(0.1)}",
        "input select face selection": f"fg:white bg:{cp.hl}",
        "input select face hovered": f"bg:{cp.bg.more(0.2)}",
        "input select face hovered selection": f"fg:white bg:{cp.hl}",
        # Dataframes
        "dataframe th": f"bg:{cp.bg.more(0.1)}",
        "dataframe row-odd td": f"bg:{cp.bg.more(0.05)}",
        # Diagnostic flags
        "diagnostic-0": "fg:ansigray",
        "diagnostic-1": "fg:ansigreen",
        "diagnostic-2": "fg:ansiblue",
        "diagnostic-3": "fg:ansiyellow",
        "diagnostic-4": "fg:ansired",
        "diagnostic-5": "fg:ansiwhite bg:ansired bold",
    }

    return Style.from_dict(style_dict)


@cache
def get_style_by_name(name: str) -> type[PygmentsStyle]:
    """Get Pygments style, caching the result."""
    return pyg_get_style_by_name(name)
